#include "math3d.h"

#ifndef __win32
#include <GL/glut.h>
#else
#include <GL\glut.h>
#endif

#include "bone3d.h"


//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
//


#define distance2d(A, B)       (sqrt( ((A)[X]-(B)[X])*((A)[X]-(B)[X])+((A)[Y]-(B)[Y])*((A)[Y]-(B)[Y]) ))


cBone::cBone()
{
  mName      = NULL;
  mTimer     = 0;
  mChildren  = NULL;
  mNext      = NULL;
}


void cBone::Animate(SCALAR delta, INTEG n, cVector *keyFrames)
{
  mTimer += delta;
  if (mTimer >= keyFrames[n-1].mV[W]) mTimer -= keyFrames[n-1].mV[W];

  INTEG i, a, b;
  for (i=0; i<n; i++)
  {
    if (keyFrames[i].mV[W] <= mTimer) {
      a = i; // interpolation start
    };
  };
  b = a + 1; // interpolation stop

  cVector d;

  d = keyFrames[b];
  d.Sub(keyFrames[a]);
  d.mV[W] -= keyFrames[a].mV[W]; // isn't sub'ed by Sub()

  SCALAR t = mTimer - keyFrames[a].mV[W];
  SCALAR f = t / d.mV[W];
  d.Mul(f);
  mAnim = keyFrames[a];
  mAnim.Add(d);
}


void cBone::ToWorld(cMatrix super, cVector length)
{
  mMatrix.Set();
  mMatrix.RotXYZ(mCenter.mV[X], mCenter.mV[Y], mCenter.mV[Z]);
  mMatrix.RotXYZ(mAnim.mV[X], mAnim.mV[Y], mAnim.mV[Z]);
  mMatrix.MovRel(length.mV[X], length.mV[Y], length.mV[Z]);
  mMatrix.Mul(super);
  
  mUJoint.Set(0, 0, 0);
  mMatrix.Apply(mUJoint);

  mLJoint = mLength;
  mMatrix.Apply(mLJoint);

  cBone *bone = mChildren;
  while (bone)
  {
    bone->ToWorld(mMatrix, mLength);
    bone = bone->mNext;
  };
}


void cBone::BaseToWorld(cMatrix super, cVector length)
{
  mMatrix.Set();
  mMatrix.RotXYZ(mCenter.mV[X], mCenter.mV[Y], mCenter.mV[Z]);
  mMatrix.MovRel(length.mV[X], length.mV[Y], length.mV[Z]);
  mMatrix.Mul(super);
  
  mUBase.Set(0, 0, 0);
  mMatrix.Apply(mUBase);

  mLBase = mLength;
  mMatrix.Apply(mLBase);

  cBone *bone = mChildren;
  while (bone)
  {
    bone->BaseToWorld(mMatrix, mLength);
    bone = bone->mNext;
  };
}


void cBone::ToWorld(cMatrix super, cMatrix rsuper, cVector length)
{
  mMatrix.Set();
  mMatrix.RotXYZ(mCenter.mV[X], mCenter.mV[Y], mCenter.mV[Z]);
  mMatrix.RotXYZ(mAnim.mV[X], mAnim.mV[Y], mAnim.mV[Z]);
  mMatrix.MovRel(length.mV[X], length.mV[Y], length.mV[Z]);
  mMatrix.Mul(super);
  
  mUJoint.Set(0, 0, 0);
  mMatrix.Apply(mUJoint);

  mLJoint = mLength;
  mMatrix.Apply(mLJoint);

  cVector r;
  r.Add(mCenter);
  r.Add(mAnim);

  mLimb.Set();
  mLimb.RotXYZ(r.mV[X], r.mV[Y], r.mV[Z]);
  mLimb.Mul(rsuper);

  cBone *bone = mChildren;
  while (bone)
  {
    bone->ToWorld(mMatrix, mLimb, mLength);
    bone = bone->mNext;
  };
}


void cBone::Draw(cMatrix camera)
{
  cMatrix c;
  c = mMatrix;
  c.Mul(camera);
  glLoadMatrixf(c.mM);

  glDisable(GL_TEXTURE_2D);
  glBegin(GL_LINES);
    glColor4f(0.0, 1.0, 0.0, 1.0);
    glVertex3f(0.0, 0.0, 0.0);
    glColor4f(0.0, 0.0, 1.0, 1.0);
    glVertex3fv(mLength.mV);
  glEnd();

  cBone *bone = mChildren;
  while (bone)
  {
    bone->Draw(camera);
    bone = bone->mNext;
  };
};


//----------------------------->


cBody::cBody()
{
  mNBones  = 0;
}


int cBody::FindBone(char *name)
{
  INTEG i, a;
  INTEG j, n, l = strlen(name);
  for (i=0; i<mNBones; i++)
  {
    if (strcmp(name, mBones[i]->mName) == 0) return i;
  };
  return -1;
}


cBone *cBody::LoadSub(FILE *f)
{
  char buf[80];
  INTEG i, n, children;
  SCALAR a, b, c;
  fscanf(f, "%s %i, %f %f %f\n", buf, &children, &a, &b, &c);
  n = FindBone(buf);
  if (n < 0) return (cBone*) NULL;
  mBones[n]->mLength.Set(a, b, c);
  cBone *bone = mBones[n];
  for (i=0; i<children; i++)
  {
    if (i == 0) {
      bone->mChildren = LoadSub(f);
      bone = bone->mChildren;
    } else {
      bone->mNext = LoadSub(f);
      bone = bone->mNext;
    };
  };
  return mBones[n];
}


void cBody::Load(FILE *f)
{
  char buf[80];
  INTEG i, n;
  fscanf(f, "%i\n", &n);
  mNBones = 0;
  for (i=0; i<n; i++)
  {
    fscanf(f, "%s\n", buf);
    mBones[mNBones] = new cBone;
    mBones[mNBones]->mName = new char[strlen(buf)+1];
    strcpy(mBones[mNBones]->mName, buf);
    mNBones++;
  };
  LoadSub(f);
}


void cBody::SaveSub(FILE *f, cBone *bone)
{
  INTEG i, children = 0;
  cBone *next = bone->mChildren;
  cVector l = bone->mLength;
  while (next) // count amount of children
  {
     children++;
     next = next->mNext;
  };
  fprintf(f, "%s %i, %f %f %f\n", bone->mName, children, l.mV[X], l.mV[Y], l.mV[Z]);
  next = bone->mChildren;
  while (next)
  {
     SaveSub(f, next);
     next = next->mNext;
  };
}


void cBody::Save(FILE *f)
{
  INTEG i;
  fprintf(f, "%i\n", mNBones);
  for (i=0; i<mNBones; i++)
  {
    fprintf(f, "%s\n", mBones[i]->mName);
  };
  SaveSub(f, mBones[0]);
}


void cBody::Scale(SCALAR f)
{
  INTEG i;
  for (i=0; i<mNBones; i++)
  {
    mBones[i]->mLength.Mul(f);
  };
}


INTEG cBody::Select(cVector pos)
{
  INTEG i, a;
  double e, d;
  a = -1;
  d = 100000000.0;
  for (i=0; i<mNBones; i++)
  {
    e = distance2d(pos.mV, mBones[i]->mLJoint.mV);
    if (e<d) {
      d = e;
      a = i;
    };
  };
  return a;
}


void cBody::Drag(cVector pos)
{
  INTEG a = Select(pos);
  if (a>=0) {
    cVector offset = pos;
    offset.Sub(mBones[a]->mLJoint);
    mBones[a]->mLength.Add(offset);
  };
}

