#ifndef __MATH3D_CPP
#define __MATH3D_CPP

// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).


#include <stdlib.h>
#include "math3d.h"

FILE                 *errlog = NULL; // error log file

// rounded to be single float

#undef PI
#define PI 3.141593

#undef M_PI
#define M_PI 3.141593

const SCALAR _180_OVER_PI = (180.0f / PI);



void normalize(VECTOR &result, VECTOR &A)
{
  SCALAR vector_length = magnitude(A);
  scalardiv( result, A, vector_length );
}


void makeNormal(VECTOR &normal, VECTOR &p0, VECTOR &p1, VECTOR &p2)
{
  VECTOR v0, v1;
  vectorsub(v0, p0, p1);
  vectorsub(v1, p1, p2);
  cross(normal, v0, v1);
}


INTEG intersection(VECTOR &normal, SCALAR d, VECTOR &p0, VECTOR &p1, VECTOR &pnt_on_plane)
{
  SCALAR d0 = calcP(normal, p0, d);
  SCALAR d1 = calcP(normal, p1, d);
  if ( xsgn(d0) == xsgn(d1) ) return 0; // sorry, no intersection!
  SCALAR dd = d0 / (d0 - d1);
  pnt_on_plane[X] = p0[X] - ((p0[X] - p1[X]) * dd);
  pnt_on_plane[Y] = p0[Y] - ((p0[Y] - p1[Y]) * dd);
  pnt_on_plane[Z] = p0[Z] - ((p0[Z] - p1[Z]) * dd);
  return 1;
}


void setMatrix(MATRIX &M, SCALAR a1, SCALAR a2, SCALAR a3, SCALAR a4,
                          SCALAR b1, SCALAR b2, SCALAR b3, SCALAR b4,
                          SCALAR c1, SCALAR c2, SCALAR c3, SCALAR c4,
                          SCALAR d1, SCALAR d2, SCALAR d3, SCALAR d4)
{
  M[ 0]=a1; M[ 1]=a2; M[ 2]=a3; M[ 3]=a4;
  M[ 4]=b1; M[ 5]=b2; M[ 6]=b3; M[ 7]=b4;
  M[ 8]=c1; M[ 9]=c2; M[10]=c3; M[11]=c4;
  M[12]=d1; M[13]=d2; M[14]=d3; M[15]=d4;
}


void setMatrix(MATRIX &M)
{
  setMatrix(M, 1,0,0,0,
               0,1,0,0,
               0,0,1,0,
               0,0,0,1);
}


void applyMatrix(MATRIX &M, VECTOR &N)
{
  VECTOR tmpN;
  tmpN[X] = (N[0]*M[0*4+0]) + (N[1]*M[1*4+0]) + (N[2]*M[2*4+0]) + (M[3*4+0]);
  tmpN[Y] = (N[0]*M[0*4+1]) + (N[1]*M[1*4+1]) + (N[2]*M[2*4+1]) + (M[3*4+1]);
  tmpN[Z] = (N[0]*M[0*4+2]) + (N[1]*M[1*4+2]) + (N[2]*M[2*4+2]) + (M[3*4+2]);
  N = tmpN;
}


void applyMatrix(MATRIX &M, VECTOR &N, VECTOR &outN)
{
  outN[X] = (N[0]*M[0*4+0]) + (N[1]*M[1*4+0]) + (N[2]*M[2*4+0]) + (M[3*4+0]);
  outN[Y] = (N[0]*M[0*4+1]) + (N[1]*M[1*4+1]) + (N[2]*M[2*4+1]) + (M[3*4+1]);
  outN[Z] = (N[0]*M[0*4+2]) + (N[1]*M[1*4+2]) + (N[2]*M[2*4+2]) + (M[3*4+2]);
}


void mulMatrix(MATRIX &M, MATRIX &N)
{
  MATRIX tmpM;
  INTEG i, j;
  for (j=0; j<4; j++) {
    for (i=0; i<4; i++) {
      tmpM[i+4*j] = (N[i+4*0]*M[0+4*j]) + (N[i+4*1]*M[1+4*j])+
                    (N[i+4*2]*M[2+4*j]) + (N[i+4*3]*M[3+4*j]);
    };
  };
  M = tmpM;
}


void addMatrix(MATRIX &M, MATRIX &N)
{
  INTEG i, j;
  for (j=0; j<4; j++)
  for (i=0; i<4; i++)
  M[i*4+j] += N[i*4+j];
}


void RXMatrix(MATRIX &M, ANGLE x)
{
  MATRIX N;
  SCALAR c = cos(x*0.017454);
  SCALAR s = sin(x*0.017454);
  setMatrix(N, 1, 0, 0, 0,
               0, c, s, 0,
               0,-s, c, 0,
               0, 0, 0, 1);
  mulMatrix(M, N);
}


void RYMatrix(MATRIX &M, ANGLE y)
{
  MATRIX N;
  SCALAR c = cos(y*0.017454);
  SCALAR s = sin(y*0.017454);
  setMatrix(N, c, 0,-s, 0,
               0, 1, 0, 0,
               s, 0, c, 0,
               0, 0, 0, 1);
  mulMatrix(M, N);
}


void RZMatrix(MATRIX &M, ANGLE z)
{
  MATRIX N;
  SCALAR c = cos(z*0.017454);
  SCALAR s = sin(z*0.017454);
  setMatrix(N, c, s, 0, 0,
              -s, c, 0, 0,
               0, 0, 1, 0,
               0, 0, 0, 1);
  mulMatrix(M, N);
}


void TMatrix(MATRIX &M, SCALAR x, SCALAR y, SCALAR z)
{
  M[12] += x;
  M[13] += y;
  M[14] += z;
}


void SMatrix(MATRIX &M, SCALAR x, SCALAR y, SCALAR z)
{
  MATRIX N;
  setMatrix(N, x, 0, 0, 0,
               0, y, 0, 0,
               0, 0, z, 0,
               0, 0, 0, 1);
  mulMatrix(M, N);
}


void RXVector(VECTOR &V, ANGLE x)
{
  MATRIX M;
  SCALAR c = cos(x*0.017454);
  SCALAR s = sin(x*0.017454);
  setMatrix(M, 1, 0, 0, 0,
               0, c, s, 0,
               0,-s, c, 0,
               0, 0, 0, 1);
  applyMatrix(M, V);
}


void RYVector(VECTOR &V, ANGLE y)
{
  MATRIX M;
  SCALAR c = cos(y*0.017454);
  SCALAR s = sin(y*0.017454);
  setMatrix(M, c, 0,-s, 0,
               0, 1, 0, 0,
               s, 0, c, 0,
               0, 0, 0, 1);
  applyMatrix(M, V);
}


void RZVector(VECTOR &V, ANGLE z)
{
  MATRIX M;
  SCALAR c = cos(z*0.017454);
  SCALAR s = sin(z*0.017454);
  setMatrix(M, c, s, 0, 0,
              -s, c, 0, 0,
               0, 0, 1, 0,
               0, 0, 0, 1);
  applyMatrix(M, V);
}


//-----------------------------


cVector::cVector()
{
  Set(0, 0, 0, 0);
}


BOOLEAN  cVector::Equal(cVector v)
{
  return ((mV[X]==v.mV[X]) && (mV[Y]==v.mV[Y]) && (mV[Z]==v.mV[Z]));
}


BOOLEAN  cVector::Unequal(cVector v)
{
  return ((mV[X]!=v.mV[X]) || (mV[Y]!=v.mV[Y]) || (mV[Z]!=v.mV[Z]));
}


void     cVector::Set(SCALAR x, SCALAR y, SCALAR z, SCALAR w)
{
  setVector(mV, x, y, z);
  mV[W] = w;
}


void     cVector::Set(SCALAR x, SCALAR y, SCALAR z)
{
  setVector(mV, x, y, z);
}


void     cVector::Set(cVector p0, cVector p1)
{
  vectorsub(mV, p0.mV, p1.mV);
}


INTEG      cVector::Perspective(SCALAR xzoom, SCALAR yzoom)
{
  if (mV[Z]<=0) return 1;
  mV[X] = xzoom * mV[X] / mV[Z];
  mV[Y] = yzoom * mV[Y] / mV[Z];
  mV[Z] = 1.0 / mV[Z];
  return 0;
}


void     cVector::Polar()
{
  cVector v;
  v.mV[X] = XYangle();
  v.mV[Y] = XZangle();
  v.mV[Z] = Len();
  *this = v;
}


void     cVector::Normalise()
{
  SCALAR vector_length = magnitude(mV);
  scalardiv(mV, mV, vector_length);
}


SCALAR   cVector::Len()
{
  return (magnitude(mV));
}


SCALAR   cVector::Dot(cVector v)
{
  return (dotprod(mV, v.mV));
}


void     cVector::Crs(cVector v)
{
  cross(mV, mV, v.mV);
}


void     cVector::Add(cVector v)
{
  vectoradd(mV, mV, v.mV);
}


void     cVector::Sub(cVector v)
{
  vectorsub(mV, mV, v.mV);
}


void     cVector::Mul(SCALAR s)
{
  scalarmul(mV, mV, s);
}


void     cVector::Div(SCALAR s)
{
  scalardiv(mV, mV, s);
}


void     cVector::Neg()
{
  negVector(mV);
}


ANGLE    cVector::XYangle()
{
  SCALAR s    = _180_OVER_PI * atan(xabs(mV[Y] / mV[X]));
  if ( mV[X] >= 0 ) {
    if ( mV[Y] >= 0 ) {
      return 90-(s);
    } else {
      return 90+(s);
    };
  } else {
    if ( mV[Y] < 0 ) {
      return 180+90-(s);
    } else {
     return 270+(s);
   };
 };
}


ANGLE    cVector::XZangle()
{
  SCALAR s    = _180_OVER_PI * atan(xabs(mV[Z] / mV[X]));
  if ( mV[X] >= 0 ) {
    if ( mV[Z] >= 0 ) {
      return 90-(s);
    } else {
      return 90+(s);
    };
  } else {
    if ( mV[Z] < 0 ) {
      return 180+90-(s);
    } else {
      return 270+(s);
    };
  };
}


ANGLE    cVector::YZangle()
{
  SCALAR s    = _180_OVER_PI * atan(xabs(mV[Z]) / mV[Y]);
  if ( mV[Y] >= 0 ) {
    if ( mV[Z] >= 0 ) {
      return 90-(s);
    } else {
      return 90+(s);
    };
  } else {
    if ( mV[Z] < 0 ) {
      return 180+90-(s);
    } else {
      return 270+(s);
    };
  };
}


INTEG cVector::FindInList(cVector *vList, INTEG n)
{
  INTEG i;
  for (i=0; i<n; i++) if (this->Equal(vList[i])) return i;
  return -1;
}


BOOLEAN cVector::InPolygon(INTEG n, cVector *vtx)
{
  INTEG i;
  cPlane edge;
  cVector mainNrm, nrm;
  makeNormal(mainNrm.mV, vtx[0].mV, vtx[1].mV, vtx[2].mV);
  mainNrm.Normalise();
  for (i=0; i<n; i++)
  {
    nrm = vtx[(i+1)%n];
    nrm.Sub(vtx[i]);
    nrm.Dot(mainNrm);
    edge.Set(nrm, vtx[i]);
    if (edge.Distance(*this) < 0) return 0;
  };
  return 1;
}


cVector cVector::GetUV(INTEG n, cVector *vtx, cVector *tex) // n textured points
{
  cMatrix   m;
  cPlane    plane;
  cVector   nrm;
  cVector   vcut[3], tcut[3];
  INTEG       i, a, b, c = 0;
  double    de, d[n];

  makeNormal(nrm.mV, vtx[0].mV, vtx[1].mV, vtx[2].mV);
  nrm.Normalise();
  m.RotX(90.0);
  m.Apply(nrm);
  plane.Set(nrm, *this);

  for (i=0; i<n; i++)
  {
    d[i] = plane.Distance(vtx[i]);
  };

  for (i=0; i<n; i++)
  {
    a = i;
    b = (i+1)%n;
    if (xsgn(d[a]) != xsgn(d[b])) {
      de = d[a] / (d[a] - d[b]);
      vcut[c].mV[X] = clif(vtx[a].mV[X], vtx[b].mV[X], de);
      vcut[c].mV[Y] = clif(vtx[a].mV[Y], vtx[b].mV[Y], de);
      vcut[c].mV[Z] = clif(vtx[a].mV[Z], vtx[b].mV[Z], de);
      tcut[c].mV[X] = clif(tex[a].mV[X], tex[b].mV[X], de);
      tcut[c].mV[Y] = clif(tex[a].mV[Y], tex[b].mV[Y], de);
      c++;
    };
    if (c >= 2) break;
  };

  a = 0;
  b = 1;
  c = 2;
  de = distance(vcut[a].mV, this->mV) / distance(vcut[a].mV, vcut[b].mV);
  vcut[c].mV[X] = clif(vcut[a].mV[X], vcut[b].mV[X], de);
  vcut[c].mV[Y] = clif(vcut[a].mV[Y], vcut[b].mV[Y], de);
  vcut[c].mV[Z] = clif(vcut[a].mV[Z], vcut[b].mV[Z], de);
  tcut[c].mV[X] = clif(tcut[a].mV[X], tcut[b].mV[X], de);
  tcut[c].mV[Y] = clif(tcut[a].mV[Y], tcut[b].mV[Y], de);

  return tcut[c];
}


void cVector::CenterOf(cVector *vList, INTEG n)
{
  INTEG i;
  Set(0, 0, 0);
  for (i=0; i<n; i++) Add(vList[i]);
  Div(n);
}


INTEG cVector::Nearest(cVector *vList, INTEG n)
{
  INTEG i, e = 0;
  SCALAR t, d = distance(mV, vList[e].mV);
  for (i=1; i<n; i++)
  {
    t = distance(mV, vList[i].mV);
    if (t < d) {
      d = t;
      e = i;
    };
  };
  return e;
}


INTEG cVector::Furthest(cVector *vList, INTEG n)
{
  INTEG i, e = 0;
  SCALAR t, d = distance(mV, vList[e].mV);
  for (i=1; i<n; i++)
  {
    t = distance(mV, vList[i].mV);
    if (t > d) {
      d = t;
      e = i;
    };
  };
  return e;
}


void cVector::Print(FILE *f)
{
  fprintf(f, "( %f, %f, %f )\n", mV[X], mV[Y], mV[Z]);
}


//-----------------------------


void cPlane::Set(cVector v, SCALAR d)
{
  mV = v;
  mV.mV[D] = d;
}


void cPlane::Set(SCALAR a, SCALAR b, SCALAR c, SCALAR d)
{
  mV.Set(a, b, c);
  mV.mV[D] = d;
}


void cPlane::Set(cVector normal, cVector point)
{
  mV = normal;
  mV.Normalise();
  mV.mV[D] = calcD(mV.mV, point.mV);
}


SCALAR cPlane::Distance(cVector point)
{
  return ( calcP(mV.mV, point.mV, mV.mV[D]) );
}


BOOLEAN cPlane::Intersect(cVector pointA, cVector pointB, cVector &pointS)
{
  SCALAR d0 = calcP(mV.mV, pointA.mV, mV.mV[D]);
  SCALAR d1 = calcP(mV.mV, pointB.mV, mV.mV[D]);
  if ( xsgn(d0) == xsgn(d1) ) return 0; // sorry, no intersection!
  SCALAR dd = d0 / (d0 - d1);
  pointS.mV[X] = pointA.mV[X] - ((pointA.mV[X] - pointB.mV[X]) * dd);
  pointS.mV[Y] = pointA.mV[Y] - ((pointA.mV[Y] - pointB.mV[Y]) * dd);
  pointS.mV[Z] = pointA.mV[Z] - ((pointA.mV[Z] - pointB.mV[Z]) * dd);
  return 1;
}


cVector cPlane::Intersect(cRay ray)
{
  cVector v = ray.mO, r = ray.mD;
  SCALAR d = calcT(mV.mV, v.mV, r.mV, mV.mV[D]);
  r.Mul(d);
  v.Add(r);
  return v;
}


//-----------------------------


cMatrix::cMatrix()
{
  Set(1,0,0,0,
      0,1,0,0,
      0,0,1,0,
      0,0,0,1);
}


void cMatrix::Set()
{
  Set(1,0,0,0,
      0,1,0,0,
      0,0,1,0,
      0,0,0,1);
}


void cMatrix::Set(SCALAR a1, SCALAR a2, SCALAR a3, SCALAR a4,
                 SCALAR b1, SCALAR b2, SCALAR b3, SCALAR b4,
                 SCALAR c1, SCALAR c2, SCALAR c3, SCALAR c4,
                 SCALAR d1, SCALAR d2, SCALAR d3, SCALAR d4)
{
  mM[ 0]=a1; mM[ 1]=a2; mM[ 2]=a3; mM[ 3]=a4;
  mM[ 4]=b1; mM[ 5]=b2; mM[ 6]=b3; mM[ 7]=b4;
  mM[ 8]=c1; mM[ 9]=c2; mM[10]=c3; mM[11]=c4;
  mM[12]=d1; mM[13]=d2; mM[14]=d3; mM[15]=d4;
}


void cMatrix::Inv()
{
  INTEG i, j, k;
  FLOAT f;
  for (i=0; i<4; i++)
  {
    f = mM [ i *4+ i ];
    mM [ i *4+ i ] = 1;
    for (j=0; j<4; j++) mM [ i *4+ j ] /= f;
    for (k=0; k<4; k++)
    {
      if ( k == i ) continue;
      f = mM [ k *4+ i ];
      mM [ k *4+ i ] = 0;
      for (j=0; j<4; j++) mM [ k *4+ j ] -= f*mM [ i *4+ j ];
    };
  };
}


void cMatrix::MovRel(SCALAR x, SCALAR y, SCALAR z)
{
  TMatrix(mM, x,y,z);
}


void cMatrix::Mul(cMatrix m)
{
  mulMatrix(mM, m.mM);
}


void cMatrix::Add(cMatrix m)
{
  addMatrix(mM, m.mM);
}


void cMatrix::Scale(SCALAR sx, SCALAR sy, SCALAR sz)
{
  SMatrix(mM, sx, sy, sz);
}


void cMatrix::RotXYZ(ANGLE a, ANGLE b, ANGLE c)
{
  RXMatrix(mM, a);
  RYMatrix(mM, b);
  RZMatrix(mM, c);
}


void cMatrix::RotX(ANGLE a)
{
  RXMatrix(mM, a);
}


void cMatrix::RotY(ANGLE a)
{
  RYMatrix(mM, a);
}


void cMatrix::RotZ(ANGLE a)
{
  RZMatrix(mM, a);
}


void cMatrix::Apply(cVector &v)
{
  cVector tmpV;
  tmpV.mV[X] = (v.mV[X]*mM[0*4+0]) + (v.mV[Y]*mM[1*4+0]) + (v.mV[Z]*mM[2*4+0]) + (mM[3*4+0]);
  tmpV.mV[Y] = (v.mV[X]*mM[0*4+1]) + (v.mV[Y]*mM[1*4+1]) + (v.mV[Z]*mM[2*4+1]) + (mM[3*4+1]);
  tmpV.mV[Z] = (v.mV[X]*mM[0*4+2]) + (v.mV[Y]*mM[1*4+2]) + (v.mV[Z]*mM[2*4+2]) + (mM[3*4+2]);
  v = tmpV;
}


void cMatrix::Apply(cVector &vOut, cVector &v)
{
  vOut.mV[X] = (v.mV[X]*mM[0*4+0]) + (v.mV[Y]*mM[1*4+0]) + (v.mV[Z]*mM[2*4+0]) + (mM[3*4+0]);
  vOut.mV[Y] = (v.mV[X]*mM[0*4+1]) + (v.mV[Y]*mM[1*4+1]) + (v.mV[Z]*mM[2*4+1]) + (mM[3*4+1]);
  vOut.mV[Z] = (v.mV[X]*mM[0*4+2]) + (v.mV[Y]*mM[1*4+2]) + (v.mV[Z]*mM[2*4+2]) + (mM[3*4+2]);
}


void cMatrix::MakeProjection(cPlane plane, cVector center)
{
  // we assume center.mV[W] == 1 to optimise
  SCALAR dot = center.mV[X] * plane.mV.mV[X] +
               center.mV[Y] * plane.mV.mV[Y] +
               center.mV[Z] * plane.mV.mV[Z] +
                             plane.mV.mV[D];

  mM[0 +  0] = dot - center.mV[X] * plane.mV.mV[X];
  mM[1 +  0] = 0.0 - center.mV[Y] * plane.mV.mV[X];
  mM[2 +  0] = 0.0 - center.mV[Z] * plane.mV.mV[X];
  mM[3 +  0] = 0.0 -               plane.mV.mV[X];

  mM[0 +  4] = 0.0 - center.mV[X] * plane.mV.mV[Y];
  mM[1 +  4] = dot - center.mV[Y] * plane.mV.mV[Y];
  mM[2 +  4] = 0.0 - center.mV[Z] * plane.mV.mV[Y];
  mM[3 +  4] = 0.0 -               plane.mV.mV[Y];

  mM[0 +  8] = 0.0 - center.mV[X] * plane.mV.mV[Z];
  mM[1 +  8] = 0.0 - center.mV[Y] * plane.mV.mV[Z];
  mM[2 +  8] = dot - center.mV[Z] * plane.mV.mV[Z];
  mM[3 +  8] = 0.0 -               plane.mV.mV[Z];

  mM[0 + 12] = 0.0 - center.mV[X] * plane.mV.mV[D];
  mM[1 + 12] = 0.0 - center.mV[Y] * plane.mV.mV[D];
  mM[2 + 12] = 0.0 - center.mV[Z] * plane.mV.mV[D];
  mM[3 + 12] = dot -               plane.mV.mV[D];
}


void cMatrix::MakeCameraRot(cVector pnt, cVector rot)
{
  Set();
  MovRel(-pnt.mV[X], -pnt.mV[Y], -pnt.mV[Z]);
  RotY(-rot.mV[Y]);
  RotX(-rot.mV[X]);
  RotZ(-rot.mV[Z]);
}


void cMatrix::MakeCameraPos(cVector pnt, cVector pos)
{
  cMatrix R;
  cVector dir;
  SCALAR xan;
  Set();
  dir.Set(pnt, pos);
  MovRel(-pnt.mV[X], -pnt.mV[Y], -pnt.mV[Z]);
  xan = dir.XZangle();
  RotY(-xan+180);
  R.RotY(-xan);
  R.Apply(dir);
  RotX(-dir.YZangle());
}


void cMatrix::MakeObjectRot(cVector pnt, cVector rot)
{
  Set();
  RotXYZ(rot.mV[X], rot.mV[Y], rot.mV[Z]);
  MovRel(pnt.mV[X], pnt.mV[Y], pnt.mV[Z]);
}


void cMatrix::MakeObjectPos(cVector pnt, cVector pos)
{
  cMatrix R;
  cVector dir;
  SCALAR xan;
  Set();
  dir.Set(pnt, pos);
  MovRel(pnt.mV[X], pnt.mV[Y], pnt.mV[Z]);
  xan = dir.XZangle();
  RotY(-xan+180);
  R.RotY(-xan);
  R.Apply(dir);
  RotX(-dir.YZangle());
}


//-----------------------------


SCALAR cSpline::GetX(SCALAR u)
{
  return spline_func(mV[0].mV[X], mV[1].mV[X], mV[2].mV[X], mV[3].mV[X], u);
}


SCALAR cSpline::GetY(SCALAR u)
{
  return spline_func(mV[0].mV[Y], mV[1].mV[Y], mV[2].mV[Y], mV[3].mV[Y], u);
}


SCALAR cSpline::GetZ(SCALAR u)
{
  return spline_func(mV[0].mV[Z], mV[1].mV[Z], mV[2].mV[Z], mV[3].mV[Z], u);
}


cVector cSpline::Get(SCALAR u)
{
  cVector v;
  v.Set(GetX(u), GetY(u), GetZ(u));
  return v;
}


//-----------------------------


void cRay::Set(cVector pointA, cVector pointB)
{
  mO = pointA;
  mD = mO;
  mD.Sub(pointB);
}


//-----------------------------


cSphere::cSphere()
{
  mR = 0;
  mR2 = 0;
}


void cSphere::Set(cVector *vList, INTEG n)
{
  mV.CenterOf(vList, n);
  mR = distance(mV.mV, vList[mV.Furthest(vList, n)].mV);
  mR2 = xsqr(mR);
}


BOOLEAN cSphere::Hit(cSphere other)
{
  SCALAR d = distance2(mV.mV, other.mV.mV);
  if (d <= (mR2 + other.mR2)) return 1;
  return 0;
}


BOOLEAN cSphere::Inside(cVector point)
{
  if (distance(mV.mV, point.mV) <= mR) return 1;
  return 0;
}


BOOLEAN cSphere::Hit(cRay ray, cVector &point)
{
  cVector H, N;
  SCALAR  a, b, c, d, s1, s2, s;
 
  H.Set(ray.mO, mV);  // origin to center

  a = xsqr(ray.mD.mV[X]) + xsqr(ray.mD.mV[Y]) + xsqr(ray.mD.mV[Z]);
  b = 2.0*(H.mV[X]*ray.mD.mV[X] + H.mV[Y]*ray.mD.mV[Y] + H.mV[Z]*ray.mD.mV[Z]);
  c = xsqr(H.mV[X]) + xsqr(H.mV[Y]) + xsqr(H.mV[Z]) - xsqr(mR);
  d = xsqr(b) - 4.0*a*c; 
 
  if(d >= 0.0) {
    s1 = -b + sqrt(d) / (2*a);
    s2 = -b - sqrt(d) / (2*a);
    if((s1 > 0) || (s2 > 0)) {   // one of em is above zero
      if (s1 <= 0) s1 = s2;      // if this one is zero then make it the other one
      if (s2 <= 0) s2 = s1;      // same here but the other way
      s = xmin(s1, s2);           // the nearest (positive)
      point = ray.mD;
      point.Mul(s);
      point.Add(ray.mO);
      N.Set(point, mV);    // center to intersection (normal of sphere)
      N.Normalise();
      return 1; 
    };
  };
  return 0; 
}


#endif

