#include <time.h>

#include "util3d.h"

#ifndef __win32
#include <GL/glut.h>
#else
#include <GL\glut.h>
#endif



cTimer::cTimer()
{
  Update();
  Update(); // yes, two times!
}


void cTimer::Update()
{
  mNewClock = clock();
  mSPF = SCALAR(mNewClock - mOldClock) / CLOCKS_PER_SEC;
  mFPS = 1.0 / mSPF;
  mOldClock = mNewClock;
}


//---------------------->


cText::cText()
{
  Reset();
}


void  cText::Reset()
{
  int i;
  for (i=0; i<255; i++) s[i] = 0;
}


int   cText::Equal(char *str)
{
  if (strcmp(s, str)==0) return 1; else return 0;
}


int   cText::Length()
{
  return (strlen(s));
}


char *cText::Lower()
{
//  return (strlwr(s));
  int i, n = Length();
  for (i=0; i<n; i++) s[i] = toupper(s[i]);
  return (s);
}


char *cText::Upper()
{
//  return (strupr(s));
  int i, n = Length();
  for (i=0; i<n; i++) s[i] = tolower(s[i]);
  return (s);
}


char *cText::Assign(char *str)
{
  return (strcpy(s, str));
}


char *cText::Append(char *str)
{
  return (strcat(s, str));
}


char *cText::Front(char *str)
{
  cText *t = new cText;
  t->Assign(str);
  t->Append(s);
  Assign(t->s);
  delete t;
  return (s);
}


char *cText::GetPtr()
{
  return s;
}


//---------------------->


cBmp::cBmp()
{
  mBPP = 0;
  mWidth = 0;
  mHeight = 0;
  mBmp = NULL;
}


void cBmp::Resize(INTEG width, INTEG height, INTEG bpp)
{
  mWidth = width;
  mHeight = height;
  mBPP = bpp;
  delete [] mBmp; mBmp = NULL;
  mBmp = new unsigned char[mWidth*mHeight*mBPP];
}


BOOLEAN cBmp::Inside(INTEG x, INTEG y)
{
  return (inarea(x, 0, mWidth-1) && inarea(y, 0, mHeight-1));
}


BOOLEAN cBmp::LoadTga(char *fname)
{
  struct {
    BYTE infosize;
    BYTE messa[1];
    BYTE typeinfo;
    BYTE messb[5];
    BYTE x[2];
    BYTE y[2];
    BYTE width[2];
    BYTE height[2];
    BYTE messc[2];
  } tgahead;

  INTEG i, j, l, w, h, bpp;
  BYTE messdata;
  FILE *tgafile;

  if( (tgafile = fopen(fname, "rb")) == NULL) {
    report("Failed to open tga file\n");
    return 1;
  };

  fread(&tgahead, sizeof(tgahead), 1, tgafile);

  if (tgahead.typeinfo!=2) {
    report("Failed to load non tga type 2 file (24bit uncompressed)\n");
    fclose(tgafile);
    return 1;
  };

  for (j=0; j<tgahead.infosize; j++) messdata = fgetc(tgafile);

  w    = (tgahead.width[1]<<8L)+(tgahead.width[0]);
  h    = (tgahead.height[1]<<8L)+(tgahead.height[0]);
  bpp  = 3;

  Resize(w, h, bpp);

  for (j=mHeight-1; j>=0; j--) {
    for (i=0; i<mWidth; i++) {
      mBmp[(i+mWidth*j)*mBPP+2] = fgetc(tgafile);
      mBmp[(i+mWidth*j)*mBPP+1] = fgetc(tgafile);
      mBmp[(i+mWidth*j)*mBPP+0] = fgetc(tgafile);
    };
  };

  fclose(tgafile);
  return 0;
}


void cBmp::MakeAlpha(cVector alphaColor)
{
  INTEG i, j, bpp = mBPP;
  BYTE r_ = (BYTE) alphaColor.mV[0];
  BYTE g_ = (BYTE) alphaColor.mV[1];
  BYTE b_ = (BYTE) alphaColor.mV[2];
  BYTE r, g, b, a;
  BYTE *buf = new BYTE[mWidth*mHeight*bpp];
  memcpy(buf, mBmp, sizeof(BYTE)*mWidth*mHeight*bpp);
  Resize(mWidth, mHeight, 4);
  for (i=0; i<mWidth; i++) {
    for (j=0; j<mHeight; j++) {
      r = buf[(i+j*mWidth)*bpp+0];
      g = buf[(i+j*mWidth)*bpp+1];
      b = buf[(i+j*mWidth)*bpp+2];
      if ( (r == r_) && (g == g_) && (b == b_) ) {
        r = 0;
        g = 0;
        b = 0;
        a = 0;
      } else {
        a = 255;
      };
      mBmp[(i+j*mWidth)*mBPP+0] = r;
      mBmp[(i+j*mWidth)*mBPP+1] = g;
      mBmp[(i+j*mWidth)*mBPP+2] = b;
      mBmp[(i+j*mWidth)*mBPP+3] = a;
    };
  };
}


void cBmp::SetPixel(INTEG x, INTEG y, INTEG z, BYTE c)
{
  if (Inside(x, y)) {
    mBmp[(x+y*mWidth)*mBPP+z] = c;
  };
}


BYTE cBmp::GetPixel(INTEG x, INTEG y, INTEG z)
{
  if (Inside(x, y)) {
    return mBmp[(x+y*mWidth)*mBPP+z];
  } else {
    return 0;
  };
}


void cBmp::CopyLayer(cBmp *srcBmp, INTEG srcZ, INTEG dstZ)
{
  INTEG w = xmin(mWidth, srcBmp->mWidth);
  INTEG h = xmin(mHeight, srcBmp->mHeight);
  INTEG i, j;
  for (j=0; j<w; j++)
  {
    for (i=0; i<h; i++)
    {
      SetPixel(i, j, dstZ, srcBmp->GetPixel(i, j, srcZ));
    };
  };
}


void cBmp::Plasma(INTEG x, INTEG y, INTEG z, INTEG w, INTEG h, BYTE c)
{
  if (w < 1) return;
  if (h < 1) return;
  Clear(x, y, z, w, h, c);
  INTEG w2 = w/2;
  INTEG h2 = h/2;
  Plasma(x,        y,  z,  w2,  h2,  c+rand()%16+8);
  Plasma(x+w2,     y,  z,  w2,  h2,  c+rand()%16+8);
  Plasma(x,     y+w2,  z,  w2,  h2,  c+rand()%16+8);
  Plasma(x+w2,  y+w2,  z,  w2,  h2,  c+rand()%16+8);
}


void cBmp::Clear(INTEG x, INTEG y, INTEG z, INTEG w, INTEG h, BYTE c)
{
  if ( (x+w > mWidth)  ||  (y+h > mHeight) ) return;
  if ( (x < 0)         ||  (y < 0)         ) return;
  INTEG i, j;
  for (j=y; j<y+h; j++)
  {
    for (i=x; i<x+w; i++)
    {
      SetPixel(i, j, z, c);
    };
  };
}


void cBmp::Clear(INTEG z, BYTE c)
{
  DWORD i;
  for (i=z; i<mWidth*mHeight*mBPP; i+=mBPP)
  {
    mBmp[i] = c;
  };
}


void cBmp::Add(INTEG z, BYTE c)
{
  INTEG e;
  DWORD i;
  for (i=z; i<mWidth*mHeight*mBPP; i+=mBPP)
  {
    e = mBmp[i] + c;
    mBmp[i] = fitarea(e, 0, 255);
  };
}


void cBmp::Mul(INTEG z, BYTE m)
{
  INTEG e;
  DWORD i;
  for (i=z; i<mWidth*mHeight*mBPP; i+=mBPP)
  {
    e = mBmp[i] * m;
    mBmp[i] = fitarea(e, 0, 255);
  };
}


void cBmp::Mul(INTEG z, SCALAR m)
{
  DWORD i;
  SCALAR e, g = xabs(m);
  for (i=z; i<mWidth*mHeight*mBPP; i+=mBPP)
  {
    e = mBmp[i] * g;
    mBmp[i] = BYTE(fitarea(e, 0, 255));
  };
}


void cBmp::Div(INTEG z, BYTE d)
{
  if (d == 0) return;
  INTEG e;
  DWORD i;
  for (i=z; i<mWidth*mHeight*mBPP; i+=mBPP)
  {
    mBmp[i] /= d;
  };
}


void cBmp::Div(INTEG z, SCALAR d)
{
  if (d == 0.0) return;
  DWORD i;
  SCALAR e, g = xabs(d);
  for (i=z; i<mWidth*mHeight*mBPP; i+=mBPP)
  {
    e = mBmp[i] / g;
    mBmp[i] = BYTE(fitarea(e, 0, 255));
  };
}



BOOLEAN cBmp::glUpload(WORD id)
{
  DWORD frmt;
  if (mBPP == 4) frmt = GL_RGBA; else frmt = GL_RGB;
  glBindTexture(GL_TEXTURE_2D, id);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  if(gluBuild2DMipmaps(GL_TEXTURE_2D, mBPP, mWidth, mHeight, (GLenum) frmt,
                       GL_UNSIGNED_BYTE, (void*) &mBmp[0])) return 1;		       
  return 0;
}


BOOLEAN cBmp::glRegister(BOOLEAN mipmap, WORD &id)
{
  glGenTextures(1, (unsigned int*) &id);
  if (this->glUpload(id)) return 1;

  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  if (!mipmap)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);

  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

  return 0;
}


