#include "control.h"

#include <stdlib.h>


//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
//


//--------------------------------


cControl::cControl()
{
  mDevice       = NULL;
  mThrottle     = 0;
  mWeapon       = 0;
}


void cControl::Process()
{
}


void cControl::Transfer()
{
  if (mDevice == NULL) return;
  mDevice->ChasisLR(mChasis.mV[Y]);  // around [Y]-axis
  mDevice->ChasisUD(mChasis.mV[X]);
  mDevice->TowerLR(mTower.mV[Y]);
  mDevice->TowerUD(mTower.mV[X]);
  mDevice->Throttle(mThrottle);
}


//--------------------------------


cAI::cAI()
{
  mDisturb = 0.5;
  ClearState();
}


void cAI::ClearState()
{
  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;
  mNav = 1;
  mStatePointer = 0;
  OverideState(WAIT, NULL);
}


void cAI::OverideState(INTEG state, cObject *object)
{
  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;
  mStateStack[mStatePointer] = state;
  mObjectStack[mStatePointer] = object;
}


void cAI::NewState(INTEG state, cObject *object)
{
  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;
  mStatePointer++;
  mStateStack[mStatePointer] = state;
  mObjectStack[mStatePointer] = object;
}


void cAI::OldState()
{
  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;
  mStatePointer--;
  if (mStatePointer<0) NewState(WAIT, NULL);
}


void cAI::Process()
{
  if (mDevice != NULL) {
    mSelf = mDevice->mPos;
  };
  if (mObjectStack[mStatePointer] != NULL) {
    mDest = mObjectStack[mStatePointer]->mPos;
  };
  switch(mStateStack[mStatePointer]) {
    case WAIT:      WaitEnemy();      break;
    case TARGET:    TargetEnemy();    break;
    case ATTACK:    AttackEnemy();    break;
    case FOLLOW:    FollowLeader();   break;
    case FLEE:      FleeEnemy();      break;
    case ROUTE:     RouteNav();       break;
    case PATROL:    PatrolArea();     break;
    default:        report("Invalid State Request!\n");     break;
  };
  Transfer();
  mTower.Set(0, 0, 0);
  mChasis.Set(0, 0, 0);
}


void cAI::WaitEnemy()
{
  mThrottle = 0;
}


void cAI::TargetEnemy()
{
  SCALAR d = distance(mSelf.mV, mDest.mV);
  if (d < 9.0)  NewState(ATTACK, mObjectStack[mStatePointer]);

  cVector dir = mSelf;    dir.Sub(mDest);
  cVector pol = dir;      pol.Polar();
  SCALAR rd;
  INTEG i;

  rd = (pol.mV[Y] - mDevice->mRot.mV[Y]);
  xmod(rd, 360);
  rd = rd - 180;

  if (rd <= 10)  mThrottle = 100;  else
  if (rd <= 20)  mThrottle =  66;  else
  if (rd <= 30)  mThrottle =  33;  else
                 mThrottle =   0;

  mChasis.mV[Y] = rd;

  rd = (pol.mV[Y] - mDevice->mRot.mV[Y] - mDevice->mTwr.mV[Y]);
  xmod(rd, 360);
  rd = rd - 180;

  mTower.mV[Y] = rd;

  pol.mV[X] -= 180;
  pol.mV[X] = xabs(pol.mV[X]);
  SCALAR rd2 = (pol.mV[X] - mDevice->mTwr.mV[X]-90);
  mTower.mV[X] = rd2;

  if (rand()%6 == 0)
  if (xabs(rd) + xabs(rd2) < 3)
  for (i=0; i<mDevice->mNWeapons; i++)
  {
    mDevice->FireWeapon(i, mObjectStack[mStatePointer]);
  };
}


void cAI::AttackEnemy()
{
  static SCALAR tf = 0;
  tf += 3;

  SCALAR d = distance(mSelf.mV, mDest.mV);
  if (d > 11.0) OldState(); // back to target

  mThrottle = 50;

  cVector dir = mSelf;    dir.Sub(mDest);
  cVector pol = dir;      pol.Polar();
  SCALAR rd;
  INTEG i;

  rd = (pol.mV[Y] - mDevice->mRot.mV[Y]);
  xmod(rd, 360);
  rd = rd - 180;

  mChasis.mV[Y] = rd;

  rd = (pol.mV[Y] - mDevice->mRot.mV[Y] - mDevice->mTwr.mV[Y]);
  xmod(rd, 360);
  rd = rd - 180;

  mTower.mV[Y] = rd;

  pol.mV[X] -= 180;
  pol.mV[X] = xabs(pol.mV[X]);
  SCALAR rd2 = (pol.mV[X] - mDevice->mTwr.mV[X]-90);
  mTower.mV[X] = rd2;

  if (rand()%6 == 0)
  if (xabs(rd) + xabs(rd2) < 3)
  for (i=0; i<mDevice->mNWeapons; i++)
  {
    mDevice->FireWeapon(i, mObjectStack[mStatePointer]);
  };

  ((cMech*)mDevice)->Pattern(tf, "nrnlln");
}


void cAI::FollowLeader()
{
}


void cAI::FleeEnemy()
{
}


void cAI::RouteNav()
{
  mDest = mObjectStack[mStatePointer]->mParts[mNav]->mPos;
  SCALAR d = distance(mSelf.mV, mDest.mV);

  if (d <= 1.0) {
    mNav++;
    if (mNav >= (mObjectStack[mStatePointer]->mNParts)) {
      mNav = 1;
      OldState();
    };
  };

  cVector dir = mSelf;    dir.Sub(mDest);
  cVector pol = dir;      pol.Polar();
  SCALAR rd;

  rd = (pol.mV[Y] - mDevice->mRot.mV[Y]);
  xmod(rd, 360);
  rd = rd - 180;

//  fprintf(stderr, "desired: %f     got: %f      turn: %f\n", pol.V[Y], mDevice->mRot.V[Y], rd);

  if (rd <= 10)  mThrottle = 100;  else
  if (rd <= 20)  mThrottle =  66;  else
  if (rd <= 30)  mThrottle =  33;  else
                 mThrottle =   0;

  mChasis.mV[Y] = rd;

  rd = (pol.mV[Y] - mDevice->mRot.mV[Y] - mDevice->mTwr.mV[Y]);
  xmod(rd, 360);
  rd = rd - 180;

  mTower.mV[Y] = rd;

//  fprintf(stderr, "Damage: %f\n", mDevice->mDamage);
}


void cAI::PatrolArea()
{
}


