#include "math3d.h"

#ifndef __win32
#include <GL/glut.h>
#endif

#ifdef __win32
#include <GL\glut.h>
#endif

#include <time.h>
#include <stdio.h>

#include "envir.h"



void cDome::DrawStuff(cMatrix camera, SCALAR x, SCALAR z, SCALAR s, DWORD tx)
{
  glColor4f(1.0, 1.0, 1.0, 0.6);
  cMatrix cm;
  cm.MovRel(x, 0.0, z);
  cm.Mul(camera);
  glLoadMatrixf(cm.mM);
  glDisable(GL_DEPTH_TEST);
  glBindTexture(GL_TEXTURE_2D, tx);
  glBegin(GL_QUADS);
    glTexCoord2f(0.0, 0.0);
    glVertex3f(+s, 0.0, -s);
    glTexCoord2f(1.0, 0.0);
    glVertex3f(-s, 0.0, -s);
    glTexCoord2f(1.0, 1.0);
    glVertex3f(-s, 0.0, +s);
    glTexCoord2f(0.0, 1.0);
    glVertex3f(+s, 0.0, +s);
  glEnd();
  glEnable(GL_DEPTH_TEST);
}


void cDome::DrawClouds(cMatrix camera, DWORD tx)
{
  clock_t c = clock();
  SCALAR e = 20.0*0.017454;
  SCALAR h = 200.0;
  SCALAR t = 0.05 * (SCALAR(c) / CLOCKS_PER_SEC);
  xmod(t, 1);
  INTEG i;
  SCALAR dx, dz;
  cVector p;
  cMatrix cm;
  cMatrix n = camera;
  n.Inv();
  n.Apply(p);
  cm.MovRel(0.0, p.mV[Y]*1.0, 0.0);
  cm.Mul(camera);
  glLoadMatrixf(cm.mM);
  glColor4fv(mLoColor.mV);
  glFogfv(GL_FOG_COLOR, mLoColor.mV);
  glDisable(GL_DEPTH_TEST);
  glBindTexture(GL_TEXTURE_2D, tx);
  dx = 1.0*t;
  dz = 1.0*t;
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(dx, dz);
  glVertex3f(0.0, h, 0.0);
  for (i=0; i<=18; i++) {
    glTexCoord2f(cos(i*e)*5.0+dx, sin(i*e)*5.0+dz);
    glVertex3f(cos(i*e)*2000.0, h, sin(i*e)*2000.0);
  };
  glEnd();
  glEnable(GL_DEPTH_TEST);
}


void cDome::DrawGround(cMatrix camera, DWORD tx)
{
  INTEG i;
  cVector p;
  cMatrix cm;
  cMatrix n = camera;
  n.Inv();
  n.Apply(p);
  cm.MovRel(p.mV[X]*1.0, 0.0, p.mV[Z]*1.0);
  cm.Mul(camera);
  SCALAR dx = p.mV[X]/5.0, dz = p.mV[Z]/5.0;
  glLoadMatrixf(cm.mM);
  glColor4fv(mLoColor.mV);
  glDisable(GL_DEPTH_TEST);
  glBindTexture(GL_TEXTURE_2D, tx);
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(dx, dz);
  glVertex3f(0.0, 0.0, 0.0);
  SCALAR e = 20.0*0.017454;
  for (i=0; i<=18; i++) {
    glTexCoord2f(cos(-i*e)*500.0+dx, sin(-i*e)*500.0+dz);
    glVertex3f(cos(-i*e)*2500.0, 0.0, sin(-i*e)*2500.0);
  };
  glEnd();
  glEnable(GL_DEPTH_TEST);
}


void cDome::DrawDome(cMatrix camera)
{
  cVector p;
  cMatrix n = camera;
  cMatrix cm;
  n.Inv();
  n.Apply(p);
  cm.MovRel(p.mV[X]*1.0, p.mV[Y]*1.0, p.mV[Z]*1.0);
  cm.Mul(camera);
  glLoadMatrixf(cm.mM);
  INTEG i, j, steps = 8;
  SCALAR y1, y2, s1, s2, f = 90.0/steps*0.017454;
  cVector color1 = mUpColor, color2 = mUpColor;
  cVector slope = mLoColor;
  slope.Sub(mUpColor);
  slope.Div((SCALAR)steps);
  glDisable(GL_DEPTH_TEST);
  s1 = sin(0*f);
  y1 = cos(0*f)*100;
  for (i=1; i<=steps; i++) {
    s2 = sin(i*f);
    y2 = cos(i*f)*100;
    color2.Add(slope);
    glBegin(GL_TRIANGLE_STRIP);
    for (j=0; j<=18; j++) {
      glColor3fv(color1.mV);
      glVertex3f(s1*cos(j*20*0.017454)*100, y1, s1*sin(j*20*0.017454)*100);
      glColor3fv(color2.mV);
      glVertex3f(s2*cos(j*20*0.017454)*100, y2, s2*sin(j*20*0.017454)*100);
    };
    glEnd();
    color1 = color2;
    y1 = y2;
    s1 = s2;
  };
  glEnable(GL_DEPTH_TEST);
}


void cDome::SetDayTime(SCALAR hour)
{
  SCALAR t = ((hour * 360.0 / 24.0) - 90.0);
  SCALAR s = sin(t*0.017454)*0.5+0.5;
  mUpColor.Set(0.0, 0.0, s*0.8+0.2, 1.0);
  mLoColor.Set(s*0.95+0.05, s*0.95+0.05, s*0.95+0.05, 1.0);
  mT = hour;
}


void cDome::SetSystemDayTime()
{
  time_t t = time(NULL);
  tm *l = localtime(&t);
  SCALAR lt = (l->tm_hour + l->tm_min / 60.0 + l->tm_sec / 3600.0);
  SetDayTime(lt);

//  SetDayTime((l->tm_sec)*24.0/60.0);
// ^- 60 seconds equals 24 hours - good for checking day-colors
}



