#define nUSE_3DFX
#define USE_GL

#ifdef __linux
#include <GL/xmesa.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <fcntl.h>
#include <unistd.h>

class cInput {
  public:
  int buttons, x, y;
};

const unsigned int JS_RETURN = sizeof(cInput);

char        *js0name = "/dev/js0",   *js1name = "/dev/js1";
int                       js0file,                 js1file;
int                     js0status,               js1status;
int                         js0xm,                   js0ym;
int                         js1xm,                   js1ym;
cInput                      js0,                     js1;
cInput                      kb0,                     kb1;


//
// Misc.
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
// aka hackCraft
//


#include "math3d.h"
#include "util3d.h"
#include "pose3d.h"
#include "conio3d.h"
#include "viewer3d.h"

#ifndef __win32
#include <GL/glut.h>
#endif

#ifdef __win32
#include <GL\glut.h>
#endif

#include "objects.h"
#include "control.h"
#include "envir.h"

#define FIRE_KEY ' '
#define JUMPJET_KEY 'j'
#define CAMERA_KEY 'c'
#define WIREFRAME_KEY 'w'
#define PAUSE_KEY 'p'
#define FXMODE_KEY 'f'



void Fatal(char *errorString)
{
#ifdef USE_3DFX
  XMesaSetFXmode(XMESA_FX_WINDOW);
#endif
  fprintf(stderr, "%s\n", errorString);
  fflush(stderr);
  exit(1);
}


BOOLEAN LoadTgaTex(char *fname, BOOLEAN transparency, BOOLEAN mipmap, WORD &TexID)
{
  cBmp *b = new cBmp;
  cVector pink;

  if (b->LoadTga(fname)) return 1;
  if (transparency) {
    pink.Set(255, 0, 255);
    b->MakeAlpha(pink);
    b->Mul(W, 0.7f); // make more transparent
  };
  return b->glRegister(mipmap, TexID);
  delete [] b;
}




cConio3d             *con[2], *cDummy;

cViewer               mainView;

cWeapon              *MGun;

cProjectile          *theShot;

cControl             *theUser;
		      
cAI                  *theAI,
                     *theAI2;

cMech                *mech,
                     *mech2,
                     *mech3;

cBuilding            *build[9];

cNavSequence         *navs;

cTree                *treeProto, *tree;

cWorld               *planet;

int                   external_camera = 0;

int                   width       = 0;
int                   height      = 0;

INTEG                 rButton     = 0;
INTEG                 mButton     = 0;
INTEG                 lButton     = 0;
SCALAR                xrSpeed     = 0;
SCALAR                yrSpeed     = 0;
SCALAR                mxpos       = 0;
SCALAR                mypos       = 0;
BOOLEAN               paused      = 0;
BOOLEAN               fullscreen  = 1;
BOOLEAN               js_disabled = 0;
BOOLEAN               wireframe   = 0;

WORD                  TexID[100];

char                  stringBuffer[512];


void Init()
{
  INTEG i, j;
  float Black[3] = { 0.0, 0.0, 0.1 };
  float lightColor[3] = { 0.5, 0.2, 1.0 };
  float fogColor[4] = { 1.0, 1.0, 1.0, 1.0 };
  float fogDensity = 0.001;

//  glDisable(GL_DITHER);
  glEnable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);

  glClearColor(0.0, 0.125, 0.0, 0.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

//  glShadeModel(GL_FLAT);
  glShadeModel(GL_SMOOTH);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDepthMask(GL_TRUE);
  glDepthFunc(GL_LEQUAL);

  glFogi(GL_FOG_MODE, GL_EXP);
  glFogfv(GL_FOG_COLOR, fogColor);
  glFogf(GL_FOG_DENSITY, fogDensity);

  glHint(GL_FOG_HINT,GL_FASTEST);

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
//  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

  if (
    LoadTgaTex("data/textures/house0.tga",    0,1, TexID[0]) |
    LoadTgaTex("data/textures/clouds0.tga",   1,1, TexID[3]) |
    LoadTgaTex("data/textures/font0.tga",     1,0, TexID[4]) |
    LoadTgaTex("data/textures/flare0.tga",    1,1, TexID[5]) |
    LoadTgaTex("data/textures/tile0.tga",     0,1, TexID[6]) |
    LoadTgaTex("data/textures/ground0.tga",   1,1, TexID[7]) |
    LoadTgaTex("data/textures/tree0.tga",     1,1, TexID[8]) 
  ) Fatal("could not load tga texture");

  con[0] = new cConio3d;
  con[0]->Resize(16, 8, 0.02);
  con[0]->SetFont(TexID[4]);
  con[0]->Insert("--------------\n");  
  con[0]->Insert("Lin Warrior 3D\n");  
  con[0]->Insert("--------------\n");  
  con[0]->Insert("              \n");  
  con[0]->Insert("3D Console 0  \n");  
  con[0]->Insert("              \n");  
  con[0]->Insert("Status: ACTIVE\n");  
  con[0]->mPose.mPos.mV[X] = +0.29;
  con[0]->mPose.mPos.mV[Y] = +0.27;
  con[0]->mPose.mRot.mV[X] = +25;
  con[0]->mPose.mRot.mV[Y] = -25;
  con[0]->mCenter.Set(-8.0*0.02, 4*0.02, 0);

  con[1] = new cConio3d;
  con[1]->Resize(16, 8, 0.02);
  con[1]->SetFont(TexID[4]);
  con[1]->Insert("--------------\n");  
  con[1]->Insert("Lin Warrior 3D\n");  
  con[1]->Insert("--------------\n");  
  con[1]->Insert("              \n");  
  con[1]->Insert("3D Console 1  \n");  
  con[1]->Insert("              \n");  
  con[1]->Insert("Status: ACTIVE\n");  
  con[1]->mPose.mPos.mV[X] = -0.29;
  con[1]->mPose.mPos.mV[Y] = +0.27;
  con[1]->mPose.mRot.mV[X] = +25;
  con[1]->mPose.mRot.mV[Y] = +25;
  con[1]->mCenter.Set(-8.0*0.02, 4*0.02, 0);

  cDummy = con[0];


  FILE *file;
  planet = new cWorld;

  mech = new cMech;
  file = fopen("data/objects/mech0.txt", "r");
  if (file==NULL) Fatal("Unable To Open Mech File...Abort.");
  if (mech->Load(file)) Fatal("Invalid Mech File...Abort");
  fclose(file);

  navs = new cNavSequence;
  file = fopen("data/objects/navs0.txt", "r");
  if (file==NULL) Fatal("Unable To Open Navs File...Abort.");
  if (navs->Load(file)) Fatal("Invalid Navs File...Abort");
  fclose(file);

  build[0] = new cBuilding;
  file = fopen("data/objects/build0.txt", "r");
  if (file==NULL) Fatal("Unable To Open Building File...Abort.");
  if (build[0]->Load(file)) Fatal("Invalid Building File...Abort");
  fclose(file);

  build[0]->mTexture = TexID[build[0]->mAttrib];
  build[0]->Link();
  for (i=1; i<9; i++) {
    build[i] = new cBuilding;
    build[i]->Copy(build[0]);
    build[i]->Link();
  };

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      if (!((i==1) && (j==1))) {
        build[i*3+j]->mPos.mV[X] = i*32-32;
        build[i*3+j]->mPos.mV[Z] = j*32-32;
	planet->AddObject(cWorld::CIVILISTIC, build[i*3+j]);
      };
    };
  };

  treeProto = new cTree;
  file = fopen("data/objects/tree0.txt", "r");
  if (file==NULL) Fatal("Unable To Open Tree File...Abort.");
  if (treeProto->Load(file)) Fatal("Invalid Tree File...Abort");
  fclose(file);

  SCALAR a, b;
  treeProto->Link();
  treeProto->mTexture = TexID[treeProto->mAttrib];
  for (i=0; i<56; i++) {
    a = rand()%360;
    b = 70 + rand() % 150;
    tree = new cTree;
    tree->Copy(treeProto);
    tree->mPos.Set(sin(a*0.017454)*b, 0, cos(a*0.017454)*b);
    tree->Link();
    planet->AddObject(cWorld::NATURAL,  tree);
  };

  theShot = new cProjectile;
  file = fopen("data/objects/shot0.txt", "r");
  if (file==NULL) Fatal("Unable To Open Building File...Abort.");
  if (theShot->Load(file)) Fatal("Invalid Building File...Abort");
  fclose(file);
  
  MGun           = new cWeapon;
  MGun->mTAmmo   = theShot;
  MGun->mNAmmo   = 1000;
  MGun->mTime    = 0;
  MGun->mHeat    = 1;
  
  theShot->mTexture   = TexID[theShot->mAttrib];
  mech->mTexture      = TexID[mech->mAttrib];
  navs->mTexture      = TexID[mech->mAttrib];

  MGun->mName = "Machine Gun";
  mech->mName = "Scorpion";
  navs->mName = "City Route";

  mech2 = new cMech;
  mech2->Copy(mech);

  mech3 = new cMech;
  mech3->Copy(mech);

  mech->Link();
  mech2->Link();
  mech3->Link();
  navs->Link();

  theUser = new cControl;
  theUser->mDevice = mech3;

  theAI = new cAI;
  theAI->mShot = theShot;
  theAI->mDevice = mech;
  theAI->NewState(cAI::TARGET, mech3);
  theAI->NewState(cAI::ROUTE, navs);

  theAI2 = new cAI;
  theAI2->mShot = theShot;
  theAI2->mDevice = mech2;
  theAI2->NewState(cAI::TARGET, mech);
//  theAI2->NewState(cAI::ROUTE, navs);
  mech2->mPos.Set(50.0, 0.0, -50.0);

  planet->mDome = new cDome;
  planet->mFireTx = TexID[5];

  planet->AddObject(cWorld::MILITARY_A,  mech);
  planet->AddObject(cWorld::MILITARY_B,  mech2);
  planet->AddObject(cWorld::MILITARY_B,  mech3);
  planet->AddObject(cWorld::NATURAL,     navs);
  planet->AddObject(cWorld::WEAPON,      MGun);

  mech->MountWeapon(cMech::LTorsor,  0);
  mech->MountWeapon(cMech::RTorsor,  0);

  mech2->MountWeapon(cMech::LLoArm,  0);
  mech2->MountWeapon(cMech::RLoArm,  0);

  mech3->MountWeapon(cMech::LTorsor, 0);
  mech3->MountWeapon(cMech::RTorsor, 0);
  mech3->MountWeapon(cMech::LLoArm,  0);
  mech3->MountWeapon(cMech::RLoArm,  0);
}


void Reshape(int w, int h)
{
  width = w;
  height = h;

#ifdef USE_3DFX
  if (width > 640) width = 640;
  if (height > 480) height = 480;
#endif

  mainView.mType = 0;
  mainView.Move(0, 0);
  mainView.Resize(width, height);

  glutPostRedisplay();
}


void Draw()
{
  if (paused) return;

  cMatrix b, a, c, d;
  static DWORD clk = 0;
  static INTEG ti = 0;
  static float tf = 0;
  INTEG i, j;

// Read in from /dev/js0 and /dev/js1 for joystick input
//

  if (!js_disabled) {  
    js0status = read(js0file, &js0, JS_RETURN);
    if (js0status != JS_RETURN) Fatal("/dev/js0 Read Error.\n");
    js1status = read(js1file, &js1, JS_RETURN);
    if (js1status != JS_RETURN) Fatal("/dev/js1 Read Error.\n");
  };


// Setup Planets Time (to determin SkyDome colors)
//

  planet->mDome->SetSystemDayTime();      // local time of "real" world
//  planet->mDome->SetDayTime(12.00);     // High Noon!!!
//  planet->mDome->SetDayTime(24.00);     // Mid Night!!!
//  planet->mDome->SetDayTime(22.00);


// Setup to prepare to Draw
//

  cDome *dome = planet->mDome;

  if (!wireframe) {
    glClearColor(dome->mLoColor.mV[X], dome->mLoColor.mV[Y], dome->mLoColor.mV[Z], 1.0);
    glFogfv(GL_FOG_COLOR, dome->mLoColor.mV);
  } else {
    cVector dark;
    dark.Set(0.0, 0.0, 0.0, 1.0);
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glFogfv(GL_FOG_COLOR, dark.mV);
  };

  glEnable(GL_FOG);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  mainView.Use();


// ------------------------->
// - D R A W  -  B E L O W ->
// ------------------------->

glEnable(GL_CULL_FACE);


  static BOOLEAN justFired=0;
  theUser->mChasis.mV[Y] = -((kb0.x+js0.x-js0xm)/50)/1.75;
  theUser->mTower.mV[X] = INTEG((-(kb0.y+js0.y-js0ym)*0.01-theUser->mDevice->mTwr.mV[X]*0.20)*3.0);
  theUser->mTower.mV[Y] = INTEG((-(js1.x-js1xm)*0.01-theUser->mDevice->mTwr.mV[Y]*0.20)*3.0);
  theUser->mThrottle = -(2*kb1.y+js1.y-js1ym)/11;
  theUser->mDevice->Jet( ((js1.buttons & 1) || (kb1.buttons & 1)) * 100.0);
  theUser->mWeapon += (js0.buttons & 2) || (kb0.buttons & 2);
  theUser->mWeapon %= theUser->mDevice->mNWeapons;
  if ((justFired==0) && ((js0.buttons & 1) || (kb0.buttons & 1))) {
    theUser->mDevice->FireWeapon(theUser->mWeapon, NULL);
    theUser->mWeapon++;
    theUser->mWeapon %= theUser->mDevice->mNWeapons;
    justFired=1;
  } else if (!(js0.buttons & 1) && !(kb0.buttons & 1)) justFired=0;
  theUser->Transfer();
  mech3->Walk(tf*20.0);
  kb0.x -= kb0.x/3;
  kb0.y -= kb0.y/3;
  kb0.buttons = 0;
  kb1.buttons = 0;

  theAI2->Process();
  mech2->Walk(tf*20.0);

  theAI->Process();
  mech->Walk(tf*20.0);


// Do some To-World-Transformations and other stuff
//

  planet->ToWorld();


// Setup camera
//

  a = mech3->mParts[cMech::Head]->mMatrix;  
  if (external_camera) {
    b.Set();
    b.MovRel(0, 0.0, -5.0);
    b.RotX(15);
    b.Mul(a);
    a = b;
  };
  a.Inv();
  a.RotY(180);

  glLoadMatrixf(a.mM); // Use it!


// Draw Lines from Mech No.1 to Mech No.2
//

  glDisable(GL_TEXTURE_2D);

  glColor4f(1.0, 0.0, 0.0, 1.0);
  glBegin(GL_LINES);
    glVertex3fv(theAI->mSelf.mV);
    glVertex3fv(theAI->mDest.mV);
  glEnd();

  glColor4f(0.0, 1.0, 0.0, 1.0);
  glBegin(GL_LINES);
    glVertex3fv(theAI2->mSelf.mV);
    glVertex3fv(theAI2->mDest.mV);
  glEnd();


// Setup the drawing scheme

  if (!wireframe) {
    glEnable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  } else {
    glDisable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  };


// Draw the environment: Ground, SkyDome, Clouds
//

  dome->DrawGround(a, TexID[7]);

  if (!wireframe) {
    glDisable(GL_FOG);
    glDisable(GL_TEXTURE_2D);
    dome->DrawDome(a);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_FOG);
    dome->DrawClouds(a, TexID[3]);
  };


// Draw the things: Mechs, Buildings, Trees, Shots, ...
//

  planet->Draw(a);


glDisable(GL_CULL_FACE);


// Draw the 3d Consoles
//

  b.Set();
  b.MovRel(0, 0, -1);

  con[0]->mPose.mRot.mV[Y] += sin(5*tf*0.017454)*0.1;
  con[0]->Draw(b);

  con[1]->mPose.mRot.mV[Y] -= sin(5*tf*0.017454)*0.1;
  con[1]->Draw(b);


// Make the drawn things visible and setup some counters
//

  glFlush();
  glutSwapBuffers();

  ti++;
  tf+=0.3;

  planet->UpdateTimers(); // calculate the time that frame took
}


void Idle()
{
  glutPostRedisplay();
}


void MouseFunc(int button, int state, int x, int y)
{
  if (button == GLUT_RIGHT_BUTTON)   rButton = state == GLUT_DOWN;
  if (button == GLUT_MIDDLE_BUTTON)  mButton = state == GLUT_DOWN;
  if (button == GLUT_LEFT_BUTTON)    lButton = state == GLUT_DOWN;
  if (x > width) x = width;
  if (y > height) y = height;
  mxpos = x;
  mypos = y;
  yrSpeed = (y-height/2.)/(height/4.);
  xrSpeed = (x-width/2.)/(width/4.);
}


void Mouse(int x, int y)
{
  if (x > width) x = width;
  if (y > height) y = height;
  mxpos = x;
  mypos = y;
  yrSpeed = (y-height/2.)/(height/4.);
  xrSpeed = (x-width/2.)/(width/4.);
}


void Key(unsigned char k, int x, int y)
{

#ifdef USE_3DFX
  if (k == FXMODE_KEY) {
    fullscreen = 1 - fullscreen;
    XMesaSetFXmode(fullscreen ? XMESA_FX_FULLSCREEN : XMESA_FX_WINDOW);
    if (fullscreen == 0) {
      cDummy->Insert("\nWINDOWED FX MODE\n");
    } else {
      cDummy->Insert("\nFULLSCREEN FX MODE\n");
    };
    if (paused) {
      paused = 0;
      Draw();
      paused = 1;
    };
  } else
#endif

  if (k == 27) {
#ifdef USE_3DFX
    XMesaSetFXmode(XMESA_FX_WINDOW);
#endif
    exit(0);
  } else

  if (k == '1') {
    kb1.y = 0;
  } else if (k == '2') {
    kb1.y = -200;
  } else if (k == '3') {
    kb1.y = -300;
  } else if (k == '4') {
    kb1.y = -400;
  } else if (k == '5') {
    kb1.y = -500;
  } else if (k == '6') {
    kb1.y = -600;
  } else if (k == '7') {
    kb1.y = -700;
  } else if (k == '8') {
    kb1.y = -800;
  } else if (k == '9') {
    kb1.y = -900;
  } else if (k == '0') {
    kb1.y = -1000;
  } else if (k == CAMERA_KEY) {
    external_camera = 1 - external_camera;
  } else if (k == FIRE_KEY) {
    kb0.buttons |= 1;
  } else if (k == JUMPJET_KEY) {
    kb1.buttons |= 1;
  } else if (k == WIREFRAME_KEY) {
    wireframe = 1 - wireframe;
  } else if (k == PAUSE_KEY) {
    if (!paused) {
      cDummy->Insert("\nPAUSED\n");
      Draw();
    } else {
      cDummy->Insert("\nSUSPEND\n");
    };
    paused = 1 - paused;
  } else {
    cDummy->InsertChar(k);
  };
}


void Special(int k, int x, int y)
{
  kb0.x = 0;
  if (k == GLUT_KEY_LEFT) kb0.x = -1000;
  if (k == GLUT_KEY_RIGHT) kb0.x = +1000;
  kb0.y = 0;
  if (k == GLUT_KEY_UP) kb0.y = -1000;
  if (k == GLUT_KEY_DOWN) kb0.y = +1000;
}


//
// GOD begins below!
//

#ifdef __win32

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  int    argc = 2;
  char  *argv[] = { "Nothing", "Nothing", "Nothing" };

#else

int main(int argc, char **argv) {

#endif

  js_disabled = 0;

  js0file = open(js0name, O_RDONLY);
  if (js0file < 0) js_disabled = 1;

  js1file = open(js1name, O_RDONLY);
  if (js1file < 0) js_disabled = 1;

  if (!js_disabled) {
    js0status = read(js0file, &js0, JS_RETURN);
    if (js0status != JS_RETURN) js_disabled = 1;

    js1status = read(js1file, &js1, JS_RETURN);
    if (js1status != JS_RETURN) js_disabled = 1;
  };
  
  js0xm = js0.x; // joystick middle is set to the current joystick position
  js0ym = js0.y;

  js1xm = js1.x;
  js1ym = js1.y;
  
  kb0.x = 0;
  kb0.y = 0;

  kb1.x = 0;
  kb1.y = 0;

  width = 640;
  height = 480;

#ifdef USE_3DFX
  putenv("MESA_GLX_FX=f");
//  putenv("MESA_GLX_FX=window");
//  putenv("SST_VGA_PASS=1");
//  putenv("SST_NOSHUTDOWN=0");
//  putenv("FX_GLIDE_NO_SPLASH=");
  width = 640;
  height = 480;
#endif

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);  
  glutInitWindowSize(width, height);
  glutInitWindowPosition(0, 0);
  glutCreateWindow("My Window");

#ifdef USE_3DFX
  glutFullScreen();
#endif

  glutDisplayFunc(*Draw);
  glutReshapeFunc(*Reshape);
  glutIdleFunc(*Idle);
  glutMouseFunc(*MouseFunc);
  glutPassiveMotionFunc(*Mouse);
  glutMotionFunc(*Mouse);
  glutKeyboardFunc(*Key);
  glutSpecialFunc(*Special);
  
  Init();
  glutMainLoop();

  if (!js_disabled) {  
    close(js1file);
    close(js0file);
  };

  return 0;
}

