#include "math3d.h"

#ifndef __win32
#include <GL/glut.h>
#endif

#ifdef __win32
#include <GL\glut.h>
#endif

#include <time.h>

#include "objects.h"
#include "envir.h"


#ifndef M_PI
#define M_PI 3.141594
#endif

cConstruct::cConstruct()
{
  mVisible     = 1;
  mNext        = NULL;
  mChildren    = NULL;
  mVolume      = 0;
  mSurface     = 0;
  mDamage      = 0;
}


void cConstruct::Calculate()
{
  mVolume      = 0;
  mSurface     = 0;
}


void cConstruct::ToWorld(cMatrix super)
{
  mMatrix.Set();
  mMatrix.MovRel(mPos.mV[X], mPos.mV[Y], mPos.mV[Z]);
  mMatrix.RotXYZ(mRot.mV[X], mRot.mV[Y], mRot.mV[Z]);
  mMatrix.Mul(super);
  mSphere.mV.Set(0, 0, 0);
  mMatrix.Apply(mSphere.mV);
  cConstruct *con = mChildren;
  while (con)
  {
    con->ToWorld(mMatrix);
    con = con->mNext;
  };
}


void cConstruct::Draw(cMatrix camera)
{
/*
  cMatrix c = mMatrix;
  c.Mul(camera);
  glLoadMatrixf(c.mM);
*/
  cConstruct *con = mChildren;
  while (con)
  {
    con->Draw(camera);
    con = con->mNext;
  };
}


//-----------------------------------


void cCriss::Calculate()
{
  mVolume = mWidth*mHeight*mLength+0.001;
  mSurface = 1.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCriss::Draw(cMatrix camera)
{
  INTEG i;
  SCALAR W2 = mWidth/2.0;
  SCALAR H2 = mHeight/2.0;
  SCALAR L2 = mLength/2.0;
  cVector p[8];
  cMatrix c = mMatrix;
  c.Mul(camera);
  glLoadMatrixf(c.mM);

  if (mVisible) {

    p[0].Set(    -W2,     +H2,      0);
    p[1].Set(    -W2,     -H2,      0);
    p[2].Set(      0,     +H2,    -L2);
    p[3].Set(      0,     -H2,    -L2);

    p[4].Set(    +W2,     +H2,      0);
    p[5].Set(    +W2,     -H2,      0);
    p[6].Set(      0,     +H2,    +L2);
    p[7].Set(      0,     -H2,    +L2);

    SCALAR wm = 1.0-mDamage/mVolume;
    if (wm<0) wm = 0;
    if (wm>1) wm = 1;
    glColor4f(1.0, wm, wm, 1.0);
    glBegin(GL_QUADS);

    glTexCoord2f(0, 0);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[2].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[6].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[2].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[1].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[1].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[5].mV);

    glEnd();
  };

  cConstruct *con = mChildren;
  while (con)
  {
    con->Draw(camera);
    con = con->mNext;
  };
}


//-----------------------------------


void cCross::Calculate()
{
  mVolume = mWidth*mHeight*mLength+0.001;
  mSurface = 1.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCross::Draw(cMatrix camera)
{
  INTEG i;
  SCALAR W2 = mWidth/2.0;
  SCALAR H2 = mHeight/2.0;
  SCALAR L2 = mLength/2.0;
  cVector p[12];
  cMatrix c = mMatrix;
  c.Mul(camera);
  glLoadMatrixf(c.mM);

  if (mVisible) {

    p[0].Set(    -W2,       0,    -L2);
    p[1].Set(    +W2,       0,    -L2);
    p[2].Set(      0,      H2,    -L2);
    p[3].Set(      0,     -H2,    -L2);

    p[4].Set(    -W2,       0,    +L2);
    p[5].Set(    +W2,       0,    +L2);
    p[6].Set(      0,      H2,    +L2);
    p[7].Set(      0,     -H2,    +L2);

    p[8].Set(    -W2,     -H2,      0);
    p[9].Set(    +W2,     -H2,      0);
    p[10].Set(   +W2,     +H2,      0);
    p[11].Set(   -W2,     -H2,      0);

    SCALAR wm = 1.0-mDamage/mVolume;
    if (wm<0) wm = 0;
    if (wm>1) wm = 1;
    glColor4f(1.0, wm, wm, 1.0);

    glBegin(GL_QUADS);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[4].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[1].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[2].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[7].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[2].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[8].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[10].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[9].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[11].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[8].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[11].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[9].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[10].mV);

    glEnd();
  };

  cConstruct *con = mChildren;
  while (con)
  {
    con->Draw(camera);
    con = con->mNext;
  };
}


//-----------------------------------


void cQuader::Calculate()
{
  mVolume = mWidth*mHeight*mLength;
  mSurface = 2.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cQuader::Draw(cMatrix camera)
{
  INTEG i;
  SCALAR W2 = mWidth/2.0;
  SCALAR H2 = mHeight/2.0;
  SCALAR L2 = mLength/2.0;
  cVector p[8];
  cMatrix c = mMatrix;
  c.Mul(camera);
  glLoadMatrixf(c.mM);

  if (mVisible) {

    p[0].Set(     0,       0,      0);
    p[1].Set(mWidth,       0,      0);
    p[2].Set(mWidth, mHeight,      0);
    p[3].Set(     0, mHeight,      0);
    p[4].Set(     0,       0, mLength);
    p[5].Set(mWidth,       0, mLength);
    p[6].Set(mWidth, mHeight, mLength);
    p[7].Set(     0, mHeight, mLength);

    for (i=0; i<8; i++)
    {
      p[i].mV[X] -= W2;
      p[i].mV[Y] -= H2;
      p[i].mV[Z] -= L2;
    };

    SCALAR wm = 1.0-mDamage/mVolume;
    if (wm<0) wm = 0;
    if (wm>1) wm = 1;
    glColor4f(1.0, wm, wm, 1.0);

    glBegin(GL_QUADS);

    glTexCoord2f(0, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[3].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[2].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[5].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[7].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[4].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[2].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[7].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[6].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[7].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[3].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[5].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[2].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[6].mV);

    glEnd();
  };

  cConstruct *con = mChildren;
  while (con)
  {
    con->Draw(camera);
    con = con->mNext;
  };
}


//-----------------------------------


void cCone::Calculate()
{
  SCALAR s = sqrt(xsqr(mHeight) + xsqr(mRadius));
  mVolume = (M_PI * xsqr(mRadius) * mHeight) / 3.0;
  mSurface = M_PI * mRadius * (mRadius + s);
  mSphere.mR = xmax(mRadius, (mHeight/2.0)); // not perfect;
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCone::Draw(cMatrix camera)
{
  INTEG i;
  cMatrix c = mMatrix;
  c.Mul(camera);
  glLoadMatrixf(c.mM);
  SCALAR wm = 1.0-mDamage/mVolume;
  if (wm<0) wm = 0;
  if (wm>1) wm = 1;
  glColor4f(1.0, wm, wm, 1.0);

  if (mVisible) {
    glBegin(GL_TRIANGLE_STRIP);
    for (i=0; i<=8; i++) {
      glTexCoord2f(i/8.0, 1);
      glVertex3f(sin(i*45.0*0.017454)*mRadius, +mHeight/2.0,  cos(i*45.0*0.017454)*mRadius);
      glTexCoord2f(i/8.0, 1);
      glVertex3f(sin(i*45.0*0.017454)*0      , -mHeight/2.0,  cos(i*45.0*0.017454)*0);
    };
    glEnd();
  };

  cConstruct *con = mChildren;
  while (con)
  {
    con->Draw(camera);
    con = con->mNext;
  };
}


//-----------------------------------


void cConeStump::Calculate()
{
  SCALAR s = sqrt(xsqr(mHeight) + xsqr(mRadiusA - mRadiusB));
  mVolume = (M_PI * mHeight * (xsqr(mRadiusA) + mRadiusA*mRadiusB + xsqr(mRadiusB) / 3.0));
  mSurface = M_PI*xsqr(mRadiusA) + M_PI*xsqr(mRadiusB) + M_PI*s*(mRadiusA + mRadiusB);
  mSphere.mR = xmax(xmax(mRadiusA, mRadiusB), (mHeight/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cConeStump::Draw(cMatrix camera)
{
  INTEG i;
  cMatrix c = mMatrix;
  c.Mul(camera);
  glLoadMatrixf(c.mM);
  SCALAR wm = 1.0-mDamage/mVolume;
  if (wm<0) wm = 0;
  if (wm>1) wm = 1;
  glColor4f(1.0, wm, wm, 1.0);

  if (mVisible) {
    glBegin(GL_TRIANGLE_STRIP);
    for (i=0; i<=8; i++) {
      glTexCoord2f(i/8.0, 1);
      glVertex3f(sin(i*45.0*0.017454)*mRadiusA, +mHeight/2.0, cos(i*45.0*0.017454)*mRadiusA);
      glTexCoord2f(i/8.0, 0);
      glVertex3f(sin(i*45.0*0.017454)*mRadiusB, -mHeight/2.0, cos(i*45.0*0.017454)*mRadiusB);
    };
    glEnd();
  };

  cConstruct *con = mChildren;
  while (con)
  {
    con->Draw(camera);
    con = con->mNext;
  };
}


//-----------------------------------


void cCylinder::Calculate()
{
  mVolume = M_PI * xsqr(mRadius) * mHeight;
  mSurface = 2.0 * M_PI * mRadius * (mRadius * mHeight);
  mSphere.mR = xmax(mRadius, (mHeight/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCylinder::Draw(cMatrix camera)
{
  INTEG i;
  cMatrix c = mMatrix;
  c.Mul(camera);
  glLoadMatrixf(c.mM);
  SCALAR wm = 1.0-mDamage/mVolume;
  if (wm<0) wm = 0;
  if (wm>1) wm = 1;
  glColor4f(1.0, wm, wm, 1.0);

  if (mVisible) {
    glBegin(GL_TRIANGLE_STRIP);
    for (i=0; i<=8; i++) {
      glTexCoord2f(i/8.0, 1);
      glVertex3f(sin(i*45.0*0.017454)*mRadius, +mHeight/2.0, cos(i*45.0*0.017454)*mRadius);
      glTexCoord2f(i/8.0, 0);
      glVertex3f(sin(i*45.0*0.017454)*mRadius, -mHeight/2.0, cos(i*45.0*0.017454)*mRadius);
    };
    glEnd();
  };

  cConstruct *con = mChildren;
  while (con)
  {
    con->Draw(camera);
    con = con->mNext;
  };
}


//-----------------------------------


cWorld::cWorld()
{
  mWind.Set(0.0, 0.0, 0.0);
  mGravity = 9.8;
  mP = 0;
  UpdateTimers();
  UpdateTimers(); // yes, two times!!
  INTEG i, j;
  for (j=0; j<6; j++) {
    mN[j] = 0;
    for (i=0; i<mMAX_OBJECTS; i++) {
      mO[j][i] = NULL;
    };
  };
}


void cWorld::UpdateTimers()
{
  mNewClock = clock();
  mSPF = SCALAR(mNewClock - mOldClock) / CLOCKS_PER_SEC;
  mFPS = 1.0 / mSPF;
  mOldClock = mNewClock;
}


void cWorld::AddObject(INTEG n, cObject *object)
{
  object->mWorld = this;
  if (mO[n][mN[n]] != NULL) delete mO[n][mN[n]];
  mO[n][mN[n]] = object;
  mN[n]++;
  if (n == PROJECTILE) {
    if (mP < mMAX_OBJECTS-1) mP = mN[n];
    mN[n] = mN[n] % mMAX_OBJECTS;
  } else {
    if (mN[n] >= mMAX_OBJECTS) mN[n] = mMAX_OBJECTS - 1;
  };
}


void cWorld::ToWorld()
{
  INTEG i, j, n;
  for (i=0; i<mP; i++) {
    ((cProjectile*)mO[4][i])->Fly();
    mO[4][i]->ToWorld();
  };
  for (j=0; j<4; j++) {
    for (i=0; i<mN[j]; i++) {
      mO[j][i]->ToWorld();
      if (j<3)
      for (n=0; n<mP; n++) if (mO[4][n]->mVolume>0)
      {
        if (mO[j][i]->Hit(mO[4][n]->mParts[1]->mSphere, 0.01)) {
//          mO[4][n]->mVolume = 0;
        };
      };
    };
  };
}


void cWorld::Draw(cMatrix c)
{
  cObject *o;
  INTEG i, j;
/*
  for (i=0; i<mP; i++) {
    o = mO[4][i];
    mDome->DrawStuff(c, o->mPos.mV[X], o->mPos.mV[Z], 1.0, mFireTx);
  };
*/
  for (j=0; j<4; j++) {
    for (i=0; i<mN[j]; i++) {
      mO[j][i]->Draw(c);
    };
  };
  for (i=0; i<mP; i++) {
    o = mO[4][i];
    o->Draw(c);
  };
}


//--------------------------------


cObject::cObject()
{
  mName         = NULL;
  mWorld        = NULL;
  mSpeed        = 0;
  mVolume       = 0;
  mTexture      = 0;
  mSurface      = 0;
  mDamage       = 0;
  mNWeapons     = 0;
  mNParts       = 1;
  mJetOn        = 0;
  mParts[0] = new cConstruct;
  mParts[0]->mType = cConstruct::Construct;
}


INTEG cObject::Hit(cSphere sphere, SCALAR damage)
{
  INTEG i, e = 0;
  for (i=1; i<mNParts; i++)
  {
    if (mParts[i]->mSphere.Hit(sphere)) {
      mParts[i]->mDamage += damage;
      e++;
    };
  };
  return e;
}


void cObject::Copy(cObject *object)
{
  INTEG i;

  mName         = object->mName;
  mNParts       = object->mNParts;
  mWorld        = object->mWorld;
  mSpeed        = object->mSpeed;
  mVolume       = object->mVolume;
  mTexture      = object->mTexture;
  mSurface      = object->mSurface;
  mDamage       = object->mDamage;
  mNParts       = object->mNParts;
  mNWeapons     = 0;
  mJetOn        = 0;

  for (i=1; i<mNParts; i++)
  {
    switch(object->mParts[i]->mType) {
      case cConstruct::Construct: {
        mParts[i] = new cConstruct;
	memcpy(mParts[i], object->mParts[i], sizeof(cConstruct));
      } break;
      case cConstruct::Cone: {
        mParts[i] = new cCone;
	memcpy(mParts[i], object->mParts[i], sizeof(cCone));
      } break;
      case cConstruct::ConeStump: {
        mParts[i] = new cConeStump;
	memcpy(mParts[i], object->mParts[i], sizeof(cConeStump));
      } break;
      case cConstruct::Cylinder: {
        mParts[i] = new cCylinder;
	memcpy(mParts[i], object->mParts[i], sizeof(cCylinder));
      } break;
      case cConstruct::Quader: {
        mParts[i] = new cQuader;
	memcpy(mParts[i], object->mParts[i], sizeof(cQuader));
      } break;
      case cConstruct::Criss: {
        mParts[i] = new cCriss;
	memcpy(mParts[i], object->mParts[i], sizeof(cCriss));
      } break;
      case cConstruct::Cross: {
        mParts[i] = new cCross;
	memcpy(mParts[i], object->mParts[i], sizeof(cCross));
      } break;
      default: break;
    };
  };
  mParts[0]->mChildren = mParts[1];
}


BOOLEAN cObject::Load(FILE *f)
{
  INTEG           t, i;
  cCone        *co;
  cConeStump   *cs;
  cCylinder    *cy;
  cQuader      *q;
  cCriss       *ci;
  cCross       *cr;
  SCALAR        x, y, z, a, b, c;

  mName         = NULL;
  mAttrib       = 0;
  mSpeed        = 0;
  mVolume       = 0;
  mSurface      = 0;
  mDamage       = 0;
  mNWeapons     = 0;
  mJetOn        = 0;

  fscanf(f, "%ld %ld\n", &t, &i);
  mAttrib = i;
  mNParts = t+1;
  for (i=1; i<mNParts; i++) {
    fscanf(f, "%ld   %f %f %f", &t, &x, &y, &z);
    switch (t) {
      case cConstruct::Construct: {
        mParts[i] = new cConstruct;
        mParts[i]->mType = cConstruct::Construct;
        mParts[i]->mPos.Set(x, y, z);
      } break;
      case cConstruct::Cone: {
        fscanf(f, "%f %f\n", &a, &b);
        mParts[i] = new cCone;
        mParts[i]->mType = cConstruct::Cone;
        mParts[i]->mPos.Set(x, y, z);
        co = (cCone*) mParts[i];
        co->mRadius = a;
        co->mHeight = b;
	mParts[i]->Calculate();
        mVolume += mParts[i]->mVolume;
      } break;
      case cConstruct::ConeStump: {
        fscanf(f, "%f %f %f\n", &a, &b, &c);
        mParts[i] = new cConeStump;
        mParts[i]->mType = cConstruct::ConeStump;
        mParts[i]->mPos.Set(x, y, z);
        cs = (cConeStump*) mParts[i];
        cs->mRadiusA = a;
        cs->mRadiusB = b;
        cs->mHeight = c;
	mParts[i]->Calculate();
        mVolume += mParts[i]->mVolume;
      } break;
      case cConstruct::Cylinder: {
        fscanf(f, "%f %f\n", &a, &b);
        mParts[i] = new cCylinder;
        mParts[i]->mType = cConstruct::Cylinder;
        mParts[i]->mPos.Set(x, y, z);
        cy = (cCylinder*) mParts[i];
        cy->mRadius = a;
        cy->mHeight = b;
	mParts[i]->Calculate();
        mVolume += mParts[i]->mVolume;
      } break;
      case cConstruct::Quader: {
        fscanf(f, "%f %f %f\n", &a, &b, &c);
        mParts[i] = new cQuader;
        mParts[i]->mType = cConstruct::Quader;
        mParts[i]->mPos.Set(x, y, z);
        q = (cQuader*) mParts[i];
        q->mWidth = a;
        q->mHeight = b;
        q->mLength = c;
	mParts[i]->Calculate();
        mVolume += mParts[i]->mVolume;
      } break;
      case cConstruct::Criss: {
        fscanf(f, "%f %f %f\n", &a, &b, &c);
        mParts[i] = new cCriss;
        mParts[i]->mType = cConstruct::Criss;
        mParts[i]->mPos.Set(x, y, z);
        ci = (cCriss*) mParts[i];
        ci->mWidth = a;
        ci->mHeight = b;
        ci->mLength = c;
	mParts[i]->Calculate();
        mVolume += mParts[i]->mVolume;
      } break;
      case cConstruct::Cross: {
        fscanf(f, "%f %f %f\n", &a, &b, &c);
        mParts[i] = new cCross;
        mParts[i]->mType = cConstruct::Cross;
        mParts[i]->mPos.Set(x, y, z);
        cr = (cCross*) mParts[i];
        cr->mWidth = a;
        cr->mHeight = b;
        cr->mLength = c;
	mParts[i]->Calculate();
        mVolume += mParts[i]->mVolume;
      } break;
      default: break;
    };
  };
  mParts[0]->mChildren = mParts[1];
  return 0;
}


void cObject::ChasisLR(SCALAR r)
{
  SCALAR e = r;
  if (e>+1) e = +1;
  if (e<-1) e = -1;
  mRot.mV[Y] += e;
  xmod(mRot.mV[Y], 360);
}


void cObject::ChasisUD(SCALAR r)
{
  SCALAR e = r;
  if (e>+1) e = +1;
  if (e<-1) e = -1;
  mRot.mV[X] += e;
  xmod(mRot.mV[X], 360);
}


void cObject::TowerLR(SCALAR r)
{
  SCALAR e = r;
  if (e>+2) e = +2;
  if (e<-2) e = -2;
  mTwr.mV[Y] += e;
  if (mTwr.mV[Y] > +75) mTwr.mV[Y] = +75;
  if (mTwr.mV[Y] < -75) mTwr.mV[Y] = -75;
}


void cObject::TowerUD(SCALAR r)
{
  SCALAR e = r;
  if (e>+2) e = +2;
  if (e<-2) e = -2;
  mTwr.mV[X] += e;
  if (mTwr.mV[X] > +75) mTwr.mV[X] = +75;
  if (mTwr.mV[X] < -75) mTwr.mV[X] = -75;
}


void cObject::Throttle(SCALAR percent)
{
  mSpeed = percent;
}


void cObject::Jet(SCALAR percent)
{
  mJetOn = percent;
}


void cObject::ToWorld()
{
  cMatrix super;
  super.RotXYZ(mRot.mV[X], mRot.mV[Y], mRot.mV[Z]);
  super.MovRel(mPos.mV[X], mPos.mV[Y], mPos.mV[Z]);
  mParts[0]->ToWorld(super);
}


void cObject::Draw(cMatrix c)
{
  if (mVolume <= 0) return;
  glBindTexture(GL_TEXTURE_2D, mTexture);
  mParts[0]->Draw(c);
}


void cObject::FireWeapon(INTEG n, cObject *target)
{
  ((cWeapon*)(mWorld->mO[cWorld::WEAPON][mTWeapons[n]]))->Fire(mParts[mMWeapons[n]], target);
}


void cObject::MountWeapon(INTEG point, INTEG type)
{
  mMWeapons[mNWeapons] = point;
  mTWeapons[mNWeapons] = type;
  mAWeapons[mNWeapons] = ((cWeapon*)(mWorld->mO[cWorld::WEAPON][type]))->mNAmmo;
  mRWeapons[mNWeapons] = 0;
  mNWeapons++;
}


void cObject::Link()
{
  INTEG i;
  for (i=1; i<mNParts-1; i++)
  {
    mParts[i]->mNext = mParts[i+1];
  };
}


//--------------------------------


cWeapon::cWeapon()
{
  mNAmmo         = 0;
  mTAmmo         = NULL;
  mTime          = 0;
  mHeat          = 0;
}


void cWeapon::Fire(cConstruct *launcher, cObject *target)
{
  cVector v;
  cMatrix w = launcher->mMatrix;
  cProjectile *shot = new cProjectile;
  shot->Copy(mTAmmo);
  shot->mGuided = mTAmmo->mGuided;
  shot->mTarget = target;
  v.Set(0, 0, 1);
  shot->Copy(mTAmmo);
  shot->Throttle(150);
  shot->mPos = launcher->mPos;
  shot->mPos.Add(v);
  shot->mTwr = shot->mPos;
  shot->mTwr.Add(v);
  w.Apply(shot->mPos);
  w.Apply(shot->mTwr);
  shot->mTwr.Sub(shot->mPos);
  v = shot->mTwr;
  v.Polar();
//  shot->mRot.Set(-mTower.mV[X], v.mV[Y], 0.0);
  mWorld->AddObject(cWorld::PROJECTILE, shot);
}


//--------------------------------


cProjectile::cProjectile()
{
  mGuided        = 0;
  mTarget        = NULL;
}


void cProjectile::Fly()
{
  SCALAR a = mSpeed / 100.0;
  SCALAR sp = a / 5.0;
/*
  const SCALAR e = 0.0174545;
  mPos.mV[X] += sp*sin(mRot.mV[Y]*e)*cos(mRot.mV[X]*e);
  mPos.mV[Y] -= sp*sin(mRot.mV[X]*e);
  mPos.mV[Z] += sp*cos(mRot.mV[Y]*e)*cos(mRot.mV[X]*e);
*/
  cVector v = mTwr;
  v.Mul(a);
  mPos.Add(v);
  if (mPos.mV[Y] < 0) mVolume = 0;
}


//--------------------------------




//--------------------------------




//-----------------------------------


BOOLEAN cMech::Hit(cRay ray, cVector &point)
{
  BOOLEAN h = 0, w = 0;
  INTEG r[15], i, j = 0;
  cVector s, p[15];
  for (i=1; i<15; i++) {
    r[i] = -1;
    if (mParts[i]->mSphere.Hit(ray, s)) {
      r[j] = i;
      p[j] = s;
      j++;
    };
  };
  if (j>0) {
    h = ray.mO.Nearest(p, j);
    mParts[r[h]]->mDamage += 0.01;
    point = p[h];
    return 1;
  } else {
    return 0;
  };
}


void cMech::Pattern(SCALAR t, char *pattern)
{
  INTEG l = strlen(pattern);
  INTEG s = INTEG(t/90.0) % l;
  char c = pattern[s];
  switch (c) {
    case 'l': ChasisLR(+5); break;
    case 'r': ChasisLR(-5); break;
    case 'n': break;
    default:  break;
  };
}


void cMech::Walk(SCALAR t)
{
  SCALAR a = mSpeed / 100.0;
  SCALAR sp = a * mWorld->mSPF * 5.0;
  SCALAR jt = mJetOn / 100.0 *mWorld->mSPF * (mWorld->mGravity + 3.0);
  const SCALAR e = 0.0174545;

  mParts[CTorsor]->mRot.mV[Y] = mTwr.mV[Y];
  mParts[CTorsor]->mRot.mV[X] = mTwr.mV[X];

  if (mPos.mV[Y] < 0.3) {
    mParts[LUpLeg]->mRot.mV[X] = (+sin(t*e)*30)*a;
    mParts[RUpLeg]->mRot.mV[X] = (-sin(t*e)*30)*a;

    mParts[LLoLeg]->mRot.mV[X] = (+sin(t*e*xsgn(a))*15+15)*xabs(a);
    mParts[RLoLeg]->mRot.mV[X] = (-sin(t*e*xsgn(a))*15+15)*xabs(a);

    mParts[LFoot]->mRot.mV[X] = 0;
    mParts[RFoot]->mRot.mV[X] = 0;
  } else {
    mParts[LUpLeg]->mRot.mV[X] = -10;
    mParts[RUpLeg]->mRot.mV[X] = -10;

    mParts[LLoLeg]->mRot.mV[X] = +20;
    mParts[RLoLeg]->mRot.mV[X] = +20;

    mParts[LFoot]->mRot.mV[X] = 0;
    mParts[RFoot]->mRot.mV[X] = 0;
  };

  mPos.mV[X] += sp*sin(mRot.mV[Y]*e)*cos(mRot.mV[X]*e);
  mPos.mV[Y] += jt - mWorld->mGravity*mWorld->mSPF;
  mPos.mV[Z] += sp*cos(mRot.mV[Y]*e)*cos(mRot.mV[X]*e);
  if (mPos.mV[Y] < 0) mPos.mV[Y] = 0;
  if (mPos.mV[Y] < 0.3) mPos.mV[Y] += sin(t*2.0*e)*0.12*a;
}


void cMech::Link()
{
  mParts[Hips]->mChildren = mParts[CTorsor];
  mParts[CTorsor]->mChildren = mParts[LTorsor];
  mParts[LTorsor]->mNext = mParts[RTorsor];
  mParts[RTorsor]->mNext = mParts[Head];
  mParts[CTorsor]->mNext = mParts[LUpLeg];
  mParts[LUpLeg]->mChildren = mParts[LLoLeg];
  mParts[LLoLeg]->mChildren = mParts[LFoot];
  mParts[LUpLeg]->mNext = mParts[RUpLeg];
  mParts[RUpLeg]->mChildren = mParts[RLoLeg];
  mParts[RLoLeg]->mChildren = mParts[RFoot];
  mParts[LTorsor]->mChildren = mParts[LUpArm];
  mParts[LUpArm]->mChildren = mParts[LLoArm];
  mParts[RTorsor]->mChildren = mParts[RUpArm];
  mParts[RUpArm]->mChildren = mParts[RLoArm];
}


//-----------------------------------


void cTank::Link()
{
  mParts[Chasis]->mChildren = mParts[Turet];
  mParts[Turet]->mNext = mParts[LChain];
  mParts[LChain]->mNext = mParts[RChain];
  mParts[Turet]->mChildren = mParts[Weapon];
}



