#ifndef OBJECTS_INC
#define OBJECTS_INC

#include <stdio.h>
#include "math3d.h"
#include "envir.h"

//-----------------------------------

class cConstruct;
  class cCone;
  class cConeStump;
  class cCylinder;
  class cQuader;
    class cCriss;
    class cCross;

class cWorld;

class cObject;
  class cWeapon;
  class cProjectile;
  class cNavSequence;
  class cBuilding;
  class cTree;
  class cMech;
  class cTank;

//-----------------------------------


class cConstruct {
  public:
  const static INTEG mMAX_PLANES = 10;
  public:
  enum               { Construct, Cone, ConeStump, Cylinder, Quader, Criss, Cross };
  public:
  INTEG              mType;
  cMatrix            mMatrix;
  BOOLEAN            mVisible;
  SCALAR             mVolume;
  SCALAR             mSurface;
  SCALAR             mDamage;
  cSphere            mSphere;
  cPlane             mPlanes[mMAX_PLANES];
  cVector            mPos;
  cVector            mRot;
  cConstruct        *mNext;
  cConstruct        *mChildren;
  public:
                     cConstruct();
  virtual void       Calculate();
  void               ToWorld(cMatrix super);
  virtual void       Draw(cMatrix camera);
};


class cCone : public cConstruct {
  public:
  SCALAR             mRadius;
  SCALAR             mHeight;
  public:
  void               Calculate();
  void               Draw(cMatrix camera);
};


class cConeStump : public cConstruct {
  public:
  SCALAR             mRadiusA;
  SCALAR             mRadiusB;
  SCALAR             mHeight;
  public:
  void               Calculate();
  void               Draw(cMatrix camera);
};


class cCylinder : public cConstruct {
  public:
  SCALAR             mRadius;
  SCALAR             mHeight;
  public:
  void               Calculate();
  void               Draw(cMatrix camera);
};


class cQuader : public cConstruct {
  public:
  SCALAR             mWidth;
  SCALAR             mHeight;
  SCALAR             mLength;
  public:
  void               Calculate();
  void               Draw(cMatrix camera);
};


class cCriss : public cQuader {
  public:
  SCALAR             mWidth;
  SCALAR             mHeight;
  SCALAR             mLength;
  public:
  void               Calculate();
  void               Draw(cMatrix camera);
};


class cCross : public cQuader {
  public:
  SCALAR             mWidth;
  SCALAR             mHeight;
  SCALAR             mLength;
  public:
  void               Calculate();
  void               Draw(cMatrix camera);
};


//-----------------------------------


class cWorld {
  public:
  const INTEG        mMAX_OBJECTS    = 64;
  const INTEG        mMAX_WEAPONS    = 32;
  public:
  enum             { MILITARY_A, MILITARY_B,
                     CIVILISTIC, NATURAL,
		     PROJECTILE, WEAPON };
  public:
  cDome             *mDome;
  
  DWORD              mFireTx;

  INTEG              mP;
  INTEG              mN[6];
  cObject           *mO[6][mMAX_OBJECTS];

  SCALAR             mTime;

  DWORD              mOldClock;
  DWORD              mNewClock;

  SCALAR             mSPF; // seconds per frame
  SCALAR             mFPS; // frames per second

  SCALAR             mGravity;
  cVector            mWind;
  public:
                     cWorld();
  void               UpdateTimers();
  void               AddObject(INTEG n, cObject *object);
  void               ToWorld();
  void               Draw(cMatrix c);
};


//-----------------------------------


class cObject {
  public:
  const INTEG          mMAX_WEAPONS = 8;
  const INTEG          mMAX_PARTS = 32;
  public:
  char              *mName;
  cWorld            *mWorld;
  INTEG              mAttrib;
  DWORD              mTexture;
  SCALAR             mVolume;
  SCALAR             mSurface;
  SCALAR             mDamage;
  SCALAR             mSpeed;
  SCALAR             mJetOn;
  cVector            mPos;
  cVector            mRot;
  cVector            mTwr;
  INTEG              mNParts;
  cConstruct        *mParts[mMAX_PARTS];
  INTEG              mNWeapons;
  INTEG              mTWeapons[mMAX_WEAPONS]; // Type Number
  INTEG              mAWeapons[mMAX_WEAPONS]; // Ammo Left
  DWORD              mRWeapons[mMAX_WEAPONS]; // Ready Time
  INTEG              mMWeapons[mMAX_WEAPONS]; // Mount Point
  public:
                     cObject();
  INTEG              Hit(cSphere sphere, SCALAR damage);
  void               Copy(cObject *object);
  BOOLEAN            Load(FILE *f);
  void               ChasisLR(SCALAR r);
  void               ChasisUD(SCALAR r);
  void               TowerLR(SCALAR r);
  void               TowerUD(SCALAR r);
  void               Throttle(SCALAR percent);
  void               Jet(SCALAR percent);
  void               ToWorld();
  void               Draw(cMatrix c);
  void               FireWeapon(INTEG n, cObject *target);
  void               MountWeapon(INTEG point, INTEG type);
  virtual void       Link();
};


class cWeapon : public cObject {
  public:
  INTEG              mNAmmo;
  cProjectile       *mTAmmo;         // Type of Ammo
  DWORD              mTime;          // mTime <= clock()   ==   ready
  SCALAR             mHeat;
  public:
                     cWeapon();
  void               Fire(cConstruct *launcher, cObject *target);
};


class cProjectile : public cObject {
  public:
  BOOLEAN            mGuided;
  cObject           *mTarget;
  public:
                     cProjectile();
  void               Fly();
};


class cNavSequence : public cObject {
  public:
};


class cBuilding : public cObject {
  public:
};


class cTree : public cObject {
  public:
};


class cMech : public cObject {
  public:
  enum               { Zero, Hips, CTorsor, LTorsor, RTorsor, Head,
                       LUpLeg, RUpLeg, LLoLeg, RLoLeg, LFoot, RFoot,
                       LUpArm, RUpArm, LLoArm, RLoArm };
  public:
  BOOLEAN            Hit(cRay ray, cVector &point);
  void               Pattern(SCALAR t, char *pattern);
  void               Walk(SCALAR t);
  void               Link();
};


class cTank : public cObject {
  public:
  enum               { Zero, Chasis, LChain, RChain, Turet, Weapon };
  public:
  void               Link();
};


#endif

